/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "JobInformationDialog.h"

#include "Application.h"
#include "BusyWidget.h"
#include "Command.h"
#include "ContextMenu.h"
#include "CustomComboBox.h"
#include "FileRecordProperties.h"
#include "FileDialog.h"
#include "GridLayout.h"
#include "GridLayoutItem.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "InformationDialog.h"
#include "JdlSyntaxHighlighter.h"
#include "JobTraceSyntaxHighlighter.h"
#include "OpenFileDialog.h"
#include "QuestionDialog.h"
#include "ServerCommand.h"
#include "Singleton.h"
#include "TextEditor.h"
#include "TreeView.h"
#include "Util.h"
#include "XmlOptions.h"

#include <QDesktopServices>
#include <QLayout>
#include <QShortcut>
#include <QTextStream>
#include <QUrl>

namespace Private
{

    enum Flag
    {
        OutputDirectory = 1<<0,
        Resubmit = 1<<1
    };

    Q_DECLARE_FLAGS( Flags, Flag );

    class ToolsPanel: public QObject, public Counter
    {

        public:

        //* constructor
        ToolsPanel( QWidget* parent ):
            QObject( parent ),
            Counter( "Private::ToolsPanel" )
        {

            // title
            title_ = new QLabel( tr( "Tools:" ), parent );
            title_->setAlignment( Qt::AlignTop|Qt::AlignRight );

            panel_ = new QWidget( parent );
            QVBoxLayout* layout = new QVBoxLayout();
            layout->setSpacing(5);
            layout->setMargin(0 );
            panel_->setLayout( layout );

            // output directory
            layout->addWidget( removeDirectoryLabel_ = new QLabel( tr( "<a href='delete'>Delete output folder ...</a>" ), panel_ ) );
            layout->addWidget( resubmitLabel_ = new QLabel( tr( "<a href='resubmit'>Resubmit job ...</a>" ), panel_ ) );

        }

        //* toggle visibility
        void setVisible( Flags flags, bool value )
        {
            Debug::Throw() << "Private::ToolsPanel::setVisible" << endl;
            if( flags & OutputDirectory ) removeDirectoryLabel_->setVisible( value );
            if( flags & Resubmit ) resubmitLabel_->setVisible( value );
            if( removeDirectoryLabel_->isVisible() || resubmitLabel_->isVisible() ) title_->setVisible( true );
            else title_->setVisible( false );
        }

        QLabel* title_;
        QWidget* panel_;
        QLabel* removeDirectoryLabel_;
        QLabel* resubmitLabel_;

    };

}

Q_DECLARE_OPERATORS_FOR_FLAGS( Private::Flags )

//______________________________________________
JobInformationDialog::JobInformationDialog( QWidget* parent ):
    CustomDialog( parent, CloseButton )
{

    Debug::Throw( "JobInformationDialog::JobInformationDialog.\n" );
    setWindowTitle( "Job Information - Gridwatch" );

    setAttribute( Qt::WA_DeleteOnClose );
    setOptionName( "JOB_INFORMATION_DIALOG" );

    // customize layout
    layout()->setMargin(0);
    layout()->setSpacing(0);
    buttonLayout().setMargin(5);

    // tabwidget
    tabWidget_ = new QTabWidget( this );
    mainLayout().addWidget( tabWidget_ );

    // general information
    QWidget *box;
    tabWidget_->addTab( box = new QWidget(), tr( "General" ) );

    QHBoxLayout* hLayout = new QHBoxLayout();
    hLayout->setMargin(5);
    hLayout->setSpacing(5);
    box->setLayout( hLayout );

    // icon
    QLabel* label = new QLabel(box);
    label->setPixmap( IconEngine::get( IconNames::DialogInformation ).pixmap( iconSize() ) );
    hLayout->addWidget( label, 0, Qt::AlignTop );

    QVBoxLayout* layout = new QVBoxLayout();
    layout->setMargin(0);
    layout->setSpacing( 5 );
    hLayout->addLayout( layout, 1 );

    GridLayout* gridLayout = new GridLayout();
    gridLayout->setSpacing( 5 );
    gridLayout->setMaxCount( 2 );
    layout->addLayout( gridLayout );

    ( idLabel_ = new GridLayoutItem( box, gridLayout, GridLayoutItem::Selectable ) )->setKey( tr( "Id:" ) );
    ( parentLabel_ = new GridLayoutItem( box, gridLayout, GridLayoutItem::Selectable ) )->setKey( tr( "Parent:" ) );
    ( userLabel_ = new GridLayoutItem( box, gridLayout, GridLayoutItem::Selectable ) )->setKey( tr( "Owner:" ) );
    ( statusLabel_ = new GridLayoutItem( box, gridLayout ) )->setKey( tr( "Status:" ) );
    ( commandLabel_ = new GridLayoutItem( box, gridLayout, GridLayoutItem::Selectable| GridLayoutItem::Elide ) )->setKey( tr( "Command:" ) );
    ( submissionTimeLabel_ = new GridLayoutItem( box, gridLayout ) )->setKey( tr( "Submitted:" ) );
    ( startTimeLabel_ = new GridLayoutItem( box, gridLayout ) )->setKey( tr( "Started:" ) );
    ( completionTimeLabel_ = new GridLayoutItem( box, gridLayout ) )->setKey( tr( "Completed:" ) );
    ( outputPathLabel_ = new GridLayoutItem( box, gridLayout, GridLayoutItem::Elide ) )->setKey( tr( "Output path:" ) );
    connect( &outputPathLabel_->label(), SIGNAL(linkActivated(QString)), SLOT(_viewOutputDirectory()) );

    outputPathLabel_->hide();

    // tools panel
    tools_ = new Private::ToolsPanel( this );
    gridLayout->addWidget( tools_->title_ );
    gridLayout->addWidget( tools_->panel_ );

    tools_->setVisible( Private::OutputDirectory|Private::Resubmit, false );

    connect( tools_->removeDirectoryLabel_, SIGNAL(linkActivated(QString)), SLOT(_deleteOutputDirectory()) );
    connect( tools_->resubmitLabel_, SIGNAL(linkActivated(QString)), SLOT(_resubmit()) );

    gridLayout->setColumnStretch( 1, 1 );
    layout->addStretch( 1 );

    // trace
    traceEditor_ = new TextEditor();
    traceEditor_->setReadOnly( true );
    new JobTraceSyntaxHighlighter( traceEditor_->document() );

    // jdl
    jdlEditor_ = new TextEditor();
    jdlEditor_->setReadOnly( true );
    jdlEditor_->setTrackAnchors( true );
    jdlSyntaxHighlighter_ = new JdlSyntaxHighlighter( jdlEditor_->document() );
    connect( jdlEditor_, SIGNAL(linkActivated(QString)), this, SLOT(_openLink(QString)) );

    // output files
    filesContainer_ = new QWidget();
    layout = new QVBoxLayout();
    layout->setMargin(5);
    layout->setSpacing(5);
    filesContainer_->setLayout( layout );

    gridLayout = new GridLayout();
    gridLayout->setMargin(0);
    gridLayout->setSpacing( 5 );
    gridLayout->setMaxCount( 2 );
    gridLayout->setColumnStretch( 1, 1 );
    layout->addLayout( gridLayout );

    // recycle path
    ( pathLabel_ = new GridLayoutItem( filesContainer_, gridLayout, GridLayoutItem::Elide ) )->setKey( tr( "Path:" ) );

    connect( &pathLabel_->label(), SIGNAL(linkActivated(QString)), SLOT(_viewOutputDirectory()) );

    // files
    layout->addWidget( list_ = new TreeView( filesContainer_ ) );
    list_->setModel( &model_ );
    list_->setSelectionMode( TreeView::SingleSelection );
    list_->setOptionName( "OUTPUT_FileS_LIST" );
    connect( list_, SIGNAL(activated(QModelIndex)), this, SLOT(_openOrSave()) );

    // reload button
    QPushButton *button = new QPushButton( IconEngine::get( IconNames::Reload ), tr( "Reload" ), this );
    connect( button, SIGNAL(clicked()), SLOT(requestUpdate()) );
    buttonLayout().insertWidget(1, button );

    // accelerators
    new QShortcut( QKeySequence::Close, this, SLOT(close()) );
    new QShortcut( QKeySequence::Refresh, this, SLOT(requestUpdate()) );

    //* actions
    _installActions();

    QMenu* menu( new ContextMenu( list_ ) );
    menu->addAction( openAction_ );
    menu->addAction( saveAsAction_ );
    connect( list_->selectionModel(), SIGNAL(selectionChanged (QItemSelection,QItemSelection)), SLOT(_updateActions()) );

    // busy widget
    busyWidget_ = new BusyWidget( tabWidget_, BusyWidget::BottomRight );
    busyWidget_->hide();

}

//_________________________________________________________________________
JobInformationDialog::~JobInformationDialog( void )
{
    jdlEditor_->deleteLater();
    traceEditor_->deleteLater();
    filesContainer_->deleteLater();
}

//_________________________________________________________________________
void JobInformationDialog::setJob( const Job& job )
{

    Debug::Throw( "JobInformationDialog::setJob.\n" );
    Q_ASSERT( job.isValid() );

    // change window title
    QString buffer;
    setWindowTitle( QString( tr( "Job %1 - Gridwatch" ) ).arg( job.id() ) );

    // assign
    job_ = job;

    // fill labels
    idLabel_->setText( QString::number(job_.id()) );
    parentLabel_->setText( job_.hasParent() ? QString::number(job_.parentId()):QString() );
    userLabel_->setText( job_.get( Job::User ) );
    statusLabel_->setText( QString( "%1 (%2)" ).arg( job_.status() ).arg( job_.get( Job::Status ) ) );

    commandLabel_->setText( job_.get( Job::Command ) );

    // assign time
    submissionTimeLabel_->setText( job.get( Job::SubmissionTime ) );
    startTimeLabel_->setText( job.get( Job::StartTime ) );
    completionTimeLabel_->setText( job.get( Job::EndTime ) );

    // clear editors
    traceEditor_->BaseEditor::clear();
    jdlEditor_->BaseEditor::clear();

    // clear output directory
    outputDirectory_.clear();

    // update tabs
    _updateTabsVisibility();

}

//_________________________________________________________________________
void JobInformationDialog::requestUpdate( void )
{

    Debug::Throw( "JobInformationDialog::requestUpdate.\n" );
    if( !job_.isValid() ) return;

    // list of pending commands
    AlienCommand::List commands;

    if( job_.isMasterJob() )
    {

        // master list
        const QString user( job_.get( Job::User ) );
        if( user == defaultUser_ ) commands.append( AlienCommand( AlienCommand::DefaultMasterList ) << user );
        else commands.append( AlienCommand( AlienCommand::MasterList ) << user );

        // job information
        commands.append( AlienCommand( AlienCommand::Trace )
            << user
            << QString::number(job_.parentId() )
            << QString::number(job_.id() ) );

        // children
        commands.append( AlienCommand( AlienCommand::ChildrenList )
            << job_.get( Job::User )
            << QString::number( job_.parentId() )
            << QString::number( job_.id() ) );

        // trace
        commands.append( AlienCommand( AlienCommand::TraceAll )
            << job_.get( Job::User )
            << QString::number(job_.parentId() )
            << QString::number(job_.id()) );

    } else {

        // children list (from parent)
        commands.append( AlienCommand( AlienCommand::ChildrenList )
            << job_.get( Job::User )
            << QString::number( -1 )
            << QString::number( job_.parentId() ) );

    }

    // jdl
    commands.append( AlienCommand( AlienCommand::TraceJdl )
        << job_.get( Job::User )
        << QString::number(job_.parentId() )
        << QString::number(job_.id()) );

    // store command ids
    commands_.clear();
    for( const auto& command:commands )
    { commands_.insert( command.id() ); }

    // make busy
    busyWidget_->start();

    // emit
    emit commandsRequested( commands );

}

//_________________________________________________________________________
void JobInformationDialog::update( const Job& job, const QString& buffer )
{

    Debug::Throw( "JobInformationDialog::update.\n" );

    // check matching
    if( job_.isValid() && job != job_ ) return;

    // reassign parent
    parentLabel_->setText( job.hasParent() ? QString::number( job.parentId() ):QString() );

    // assign time and trace
    submissionTimeLabel_->setText( job.get( Job::SubmissionTime ) );
    startTimeLabel_->setText( job.get( Job::StartTime ) );
    completionTimeLabel_->setText( job.get( Job::EndTime ) );

    // log
    if( !buffer.isEmpty() )
    { traceEditor_->setText( buffer ); }

    _updateTabsVisibility();

}


//_________________________________________________________________________
void JobInformationDialog::updateStatus( const Job& job )
{

    Debug::Throw( "JobInformationDialog::updateStatus.\n" );

    // check matching
    if( job_.isValid() && job != job_ ) return;

    // copy relevant information to stored job
    job_.setParentId( job.parentId() );
    job_.assign( Job::User, job.get( Job::User ) );
    job_.assign( Job::Status, job.get( Job::Status ) );
    job_.assign( Job::Command, job.get( Job::Command ) );

    // update labels
    idLabel_->setText( QString::number(job_.id()) );
    parentLabel_->setText( job_.hasParent() ? QString::number( job_.parentId() ):QString() );
    userLabel_->setText( job_.get( Job::User ) );
    statusLabel_->setText( QString( "%1 (%2)" ).arg( job_.status() ).arg( job_.get( Job::Status ) ) );
    commandLabel_->setText( job.get( Job::Command ) );

    _updateTabsVisibility();

}

//_________________________________________________________________________
void JobInformationDialog::updateStatus( QString user, const JobSet& jobs )
{

    Debug::Throw( "JobInformationDialog::updateStatus.\n" );

    // check job
    if( job_.isValid() && job_.get( Job::User ) != user ) return;

    JobSet::const_iterator jobIter( jobs.find( job_ ) );
    if( jobIter != jobs.end() ) updateStatus( *jobIter );

}

//_________________________________________________________________________
void JobInformationDialog::updateStatus( const Job& parent, const JobSet& jobs )
{

    Debug::Throw( "JobInformationDialog::updateStatus.\n" );

    // check job
    if( job_.isValid() && job_.parentId() != parent.id() ) return;

    JobSet::const_iterator jobIter( jobs.find( job_ ) );
    if( jobIter != jobs.end() ) updateStatus( *jobIter );

}

//_________________________________________________________________________
void JobInformationDialog::updateJdl( const Job& job, const QString& buffer )
{
    Debug::Throw( "JobInformationDialog::updateJdl.\n" );

    // check matching
    if( job_.isValid() && job != job_ ) return;

    // check buffer
    if( buffer.isEmpty() ) return;

    // assign to editor
    jdlEditor_->setText( buffer );
    jdlSyntaxHighlighter_->updateLinks();

    _updateTabsVisibility();

    // try load output directory
    QRegExp outputDirRegExp( "OutputDir = \"(\\S+)\";" );
    if( outputDirRegExp.indexIn( buffer ) >= 0 )
    {
        outputDirectory_ = outputDirRegExp.cap(1);

        // truncate path
        if( job_.isMasterJob() && outputDirectory_.endsWith( '#' ) )
        { outputDirectory_ = File( outputDirectory_ ).path(); }

        outputPathLabel_->setText( QString( "<a href='view'>%1</a>" ).arg( outputDirectory_ ) );
        pathLabel_->setText( QString( "<a href='view'>%1</a>" ).arg( outputDirectory_ ) );
    }

    if( job_.isCompleted() && job_.hasParent() && !outputDirectory_.isEmpty() )
    { _requestFiles( outputDirectory_ ); }

    _updateTabsVisibility();

}

//_________________________________________________________________________
void JobInformationDialog::updateFiles( const Job& job, const QString& path, const FileRecord::List& files )
{
    Debug::Throw( "JobInformationDialog::updateFiles.\n" );

    // check matching
    if( job_.isValid() && job != job_ ) return;

    Model::List records;
    for( const auto& record:files )
    { records << record; }

    model_.set( records );
    list_->resizeColumns();

    _updateTabsVisibility();

}

//_________________________________________________________________________
void JobInformationDialog::commandCompleted( const AlienCommand& command )
{
    Debug::Throw( "JobInformationDialog::commandCompleted.\n" );
    commands_.remove( command.id() );
    if( commands_.isEmpty() )  busyWidget_->stop();
}

//_________________________________________________________________________
void JobInformationDialog::_openOrSave( OpenFileDialog::Commands mode )
{
    Debug::Throw( "JobInformationDialog::_openOrSave.\n" );
    QModelIndex index( list_->selectionModel()->currentIndex() );
    if( !index.isValid() ) return;

    // get remote filename
    File remote( model_.get(index).file() );

    OpenFileDialog dialog( this, remote, mode );
    dialog.setWindowTitle( tr( "Open file - Gridwatch" ) );

    // set command from Options
    if( mode == OpenFileDialog::Any )
    { dialog.setCommand( (OpenFileDialog::Command) XmlOptions::get().get<int>( "OPEN_COMMAND" ) ); }

    if( dialog.centerOnParent().exec() == QDialog::Rejected ) return;

    // save command
    if( mode == OpenFileDialog::Any )
    { XmlOptions::get().set<int>( "OPEN_COMMAND", dialog.command() ); }

    // take action depending on dialog
    if( dialog.command() == OpenFileDialog::SaveAs ) _saveAs( remote );
    else {

        // retrieve application from combobox and add as options
        if( !dialog.isOpenWithCommandValid() || (dialog.openWithCommand().isEmpty() && !dialog.isOpenWithCommandDefault() ) )
        {
            InformationDialog( this, tr( "No command specified to open the selected files. <Open> canceled." ) ).exec();
            return;
        }

        // temporary path
        emit scratchFileCreated( Util::tmp() + "/gridwatch" );
        emit scratchFileCreated( Util::tmp() + "/gridwatch/" + Util::user() );

        // generate temporary filename
        QString path;
        QTextStream( &path )
            << Util::tmp() << "/gridwatch/"
            << Util::user()
            << "/alien-job-"
            << job_.id();

        File local( remote.localName().addPath( path ) );
        Debug::Throw() << "JobInformationDialog::_openOrSave - remote: " << remote << endl;
        Debug::Throw() << "JobInformationDialog::_openOrSave - local: " << local << endl;

        // make sure local path exists
        {
            File path( local.path() );
            if( !( path.exists() || path.createDirectory() ) ) return;
            emit scratchFileCreated( local.path() );
        }

        // register local file to scratch manager
        emit scratchFileCreated( local );

        // run command
        if( !dialog.isOpenWithCommandDefault() )
        {
            const QString command( dialog.openWithCommand() );
            emit commandRequested( AlienCommand( AlienCommand::Open )
                << remote
                << local
                << command );
         } else {

            emit commandRequested( AlienCommand( AlienCommand::Open )
                << remote
                << local );

        }

    }

}

//_________________________________________________________________________
void JobInformationDialog::_open( void )
{
    Debug::Throw( "JobInformationDialog::_open.\n" );
    return _openOrSave( OpenFileDialog::OpenWith );
}


//_________________________________________________________________________
void JobInformationDialog::_saveAs( File remote )
{
    Debug::Throw( "JobInformationDialog::_saveAs.\n" );
    if( remote.isEmpty() )
    {

        QModelIndex index( list_->selectionModel()->currentIndex() );
        if( !index.isValid() ) return;
        remote = model_.get(index).file();
    }

    // create file dialog
    FileDialog dialog( this );
    dialog.setFileMode( QFileDialog::AnyFile );
    dialog.setAcceptMode( QFileDialog::AcceptSave );
    dialog.selectFile( remote.localName() );
    File local( dialog.getFile() );
    if( local.isNull() ) return;
    else local = local.expand();

    // check if file is directory
    if( local.isDirectory() )
    {
        InformationDialog( this, QString( tr( "File \"%1\" is a directory. <Save As> canceled." ) ).arg( local ) ).exec();
        return;
    }

    // run command
    emit commandRequested( AlienCommand( AlienCommand::Copy )
        << remote
        << local );

}

//__________________________________________________
void JobInformationDialog::_deleteOutputDirectory( void )
{

    Debug::Throw( "JobInformationDialog::_deleteOutputDirectory.\n" );

    // check user
    if( job_.get( Job::User ) != defaultUser_ ) return;

    // check output directory
    if( outputDirectory_.isEmpty() ) return;

    // confirmation dialog
    QuestionDialog dialog( this, QString( tr( "Permanently delete job's output folder '%1' ?" ).arg( outputDirectory_ ) ) );
    dialog.okButton().setText( tr( "Remove" ) );
    dialog.okButton().setIcon( IconEngine::get( IconNames::Delete ) );
    if( !dialog.exec() ) return;

    // submit command
    emit commandRequested( AlienCommand( AlienCommand::RemoveDirectory ) << outputDirectory_ );

}

//__________________________________________________
void JobInformationDialog::_viewOutputDirectory( void )
{
    Debug::Throw( "JobInformationDialog::_viewOutputDirectory.\n" );

    // define anchor
    const QString anchor( QString( "alien://%1" ).arg( outputDirectory_ ) );

    // try using desktop services
    if( !QDesktopServices::openUrl( QUrl::fromEncoded( anchor.toLatin1() ) ) )
    {

        // send server command
        Server::ApplicationId id( "Gridftp" );
        Server::ServerCommand command( id, Server::ServerCommand::Request );
        command.setArguments( QStringList() << "Gridftp" << anchor );
        static_cast<Application*>(Singleton::get().application())->sendServerCommand( command );

    }

}

//__________________________________________________
void JobInformationDialog::_resubmit( void )
{

    Debug::Throw() << "JobInformationDialog::_resubmit" << endl;

    if( !job_.isValid() ) return;
    if( job_.isMasterJob() ) return;

    // check user
    const QString user( job_.get( Job::User ) );
    if( user != defaultUser_ ) return;

    if( !job_.isFailed() )
    {
        QuestionDialog dialog( this, QString( tr( "Job %1 is not in a failed state.\nResubmit anyway ?" ) ).arg( job_.id() ) );
        dialog.okButton().setText( tr( "Resubmit" ) );
        dialog.okButton().setIcon( IconEngine::get( IconNames::Resubmit ) );
        if( !dialog.exec() ) return;
    }

    emit commandRequested( AlienCommand( AlienCommand::Resubmit )
        << user
        << QString::number(job_.parentId())
        << QString::number(job_.id() ) );

    requestUpdate();

}

//_____________________________________________
void JobInformationDialog::_openLink( QString anchor )
{

    Debug::Throw( "MainWindow::_openLink.\n" );
    if( anchor.isEmpty() ) return;

    if( !QDesktopServices::openUrl( QUrl::fromEncoded( anchor.toLatin1() ) ) )
    {
        // send server command
        Server::ApplicationId id( "Gridftp" );
        Server::ServerCommand command( id, Server::ServerCommand::Request );
        command.setArguments( QStringList() << "Gridftp" << anchor );
        static_cast<Application*>(Singleton::get().application())->sendServerCommand( command );

    }

}

//__________________________________________________
void JobInformationDialog::_updateActions( void )
{
    const bool hasSelection( !list_->selectionModel()->selectedIndexes().isEmpty() );
    openAction_->setEnabled( hasSelection );
    saveAsAction_->setEnabled( hasSelection );
}

//________________________________________________________________________
void JobInformationDialog::_installActions( void )
{
    Debug::Throw( "JobInformationDialog::_installActions.\n" );
    addAction( openAction_ = new QAction( IconEngine::get( IconNames::Open ), tr( "Open With..." ), this ) );
    connect( openAction_, SIGNAL(triggered()), SLOT(_open()) );
    openAction_->setShortcut( QKeySequence::Open );
    openAction_->setEnabled( false );

    addAction( saveAsAction_ = new QAction( IconEngine::get( IconNames::SaveAs ), tr( "Save as..." ), this ) );
    connect( saveAsAction_, SIGNAL(triggered()), SLOT(_saveAs()) );
    saveAsAction_->setShortcut( QKeySequence::SaveAs );
    saveAsAction_->setEnabled( false );
}

//________________________________________________________________________
void JobInformationDialog::_requestFiles( const QString& path )
{
    Debug::Throw( "JobInformationDialog::_requestTrace.\n" );
    emit commandRequested( AlienCommand( AlienCommand::JobFiles )
        << job_.get( Job::User )
        << QString::number(job_.parentId() )
        << QString::number(job_.id())
        << path );
    return;
}

//________________________________________________________
void JobInformationDialog::_updateTabsVisibility( void )
{

    {
        // trace
        const int tabIndex( tabWidget_->indexOf( traceEditor_ ) );
        if( job_.isMasterJob() && !traceEditor_->toPlainText().isEmpty() && tabIndex < 0 ) tabWidget_->insertTab( 1, traceEditor_, tr( "Trace" ) );
        else if( ( job_.hasParent() || traceEditor_->toPlainText().isEmpty() ) && tabIndex >= 0 ) tabWidget_->removeTab( tabIndex );
    }

    {
        // jdl
        const int tabIndex( tabWidget_->indexOf( jdlEditor_ ) );
        if( !jdlEditor_->toPlainText().isEmpty() && tabIndex < 0 ) tabWidget_->insertTab( 1, jdlEditor_, tr( "JDL" ) );
        else if( jdlEditor_->toPlainText().isEmpty() && tabIndex >= 0 ) tabWidget_->removeTab( tabIndex );
    }

    {
        // show/hide output files tab
        const int tabIndex( tabWidget_->indexOf( filesContainer_ ) );
        if( job_.isCompleted() && job_.hasParent() && !outputDirectory_.isEmpty() && tabIndex < 0 ) tabWidget_->addTab( filesContainer_, tr( "Output Files" ) );
        else if( (job_.isMasterJob() || !job_.isCompleted() || outputDirectory_.isEmpty() ) && tabIndex >= 0 ) tabWidget_->removeTab( tabIndex );
    }

    {
        // tools
        tools_->setVisible( Private::OutputDirectory, job_.hasParent() && !job_.isRunning() && !outputDirectory_.isEmpty() );
        tools_->setVisible( Private::Resubmit, job_.hasParent() );
    }

}

//_________________________________________________________________________
QString JobInformationDialog::_recyclePath( void ) const
{
    // get user and check
    const QString user( job_.get( Job::User ) );
    Q_ASSERT( !user.isEmpty() );

    QString out;
    QTextStream( &out ) << "/alice/cern.ch/user/" << user[0] << "/" << user << "/recycle/alien-job-" << job_.id() << "/";
    return out;
}

//_________________________________________________________________________
const QString JobInformationDialog::Model::columnTitles_[ JobInformationDialog::Model::nColumns ] =
{
    tr( "Permissions" ),
    tr( "Owner" ),
    tr( "Group" ),
    tr( "Size" ),
    tr( "Date" ),
    tr( "File Name" )
};

//__________________________________________________________________
QVariant JobInformationDialog::Model::data( const QModelIndex& index, int role ) const
{

    // check index
    if( !contains( index ) ) return QVariant();

    // retrieve associated file info
    const FileRecord& record( get(index) );

    // return text associated to file and column
    if( role == Qt::DisplayRole ) {

        switch( index.column() )
        {
            case Permissions: return record.property( FileRecordProperties::Permissions );
            case User: return record.property( FileRecordProperties::User );
            case Group: return record.property( FileRecordProperties::Group );
            case Size: return File::sizeString( record.property( FileRecordProperties::Size ).toInt() );
            case Creation: return record.property( FileRecordProperties::Creation );
            case Filename: return record.file().localName();
            default: return QVariant();
        }
    }

    return QVariant();

}

//__________________________________________________________________
QVariant JobInformationDialog::Model::headerData(int section, Qt::Orientation orientation, int role) const
{

    if(
        orientation == Qt::Horizontal &&
        role == Qt::DisplayRole &&
        section >= 0 &&
        section < nColumns )
    { return columnTitles_[section]; }

    // return empty
    return QVariant();

}

//____________________________________________________________
void JobInformationDialog::Model::_sort( int column, Qt::SortOrder order )
{ std::sort( _get().begin(), _get().end(), SortFTor( (ColumnType) column, order ) ); }

//________________________________________________________
bool JobInformationDialog::Model::SortFTor::operator () ( FileRecord first, FileRecord second ) const
{

    if( order_ == Qt::DescendingOrder ) std::swap( first, second );

    switch( type_ )
    {
        case Permissions: return first.property( FileRecordProperties::Permissions ) < first.property( FileRecordProperties::Permissions );
        case User: return first.property( FileRecordProperties::User ) < first.property( FileRecordProperties::User );
        case Group: return first.property( FileRecordProperties::Group ) < first.property( FileRecordProperties::Group );
        case Size: return first.property( FileRecordProperties::Size ).toInt() < first.property( FileRecordProperties::Size ).toInt();
        case Creation: return first.property( FileRecordProperties::Creation ) < first.property( FileRecordProperties::Creation );
        case Filename: return first.file().localName() < first.file().localName();
        default: return true;
    }

}
