#ifndef JobInformationDialog_h
#define JobInformationDialog_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AlienCommand.h"
#include "Counter.h"
#include "CustomDialog.h"
#include "FileRecord.h"
#include "JdlSyntaxHighlighter.h"
#include "Job.h"
#include "JobSet.h"
#include "Key.h"
#include "ListModel.h"
#include "OpenFileDialog.h"

#include <QLabel>
#include <QTabWidget>

class BusyWidget;
class GridLayoutItem;
class TextEditor;
class TreeView;

namespace Private
{

    class ToolsPanel;

}

class JobInformationDialog: public CustomDialog, public Base::Key
{

    Q_OBJECT

    public:

    //* constructor
    JobInformationDialog( QWidget* );

    //* destructor
    virtual ~JobInformationDialog( void );

    //*@name accessors
    //@{

    //* job
    const Job& job( void ) const
    { return job_; }

    //@}

    //*@name modifiers
    //@{

    //* set default user
    void setDefaultUser( const QString& user )
    { defaultUser_ = user; }

    //* set job
    void setJob( const Job& );

    //@}

    Q_SIGNALS:

    //* request commands
    void commandsRequested( const AlienCommand::List& );

    //* request command
    void commandRequested( const AlienCommand& );

    //* scratch file created
    void scratchFileCreated( const File& );

    public Q_SLOTS:

    //* request update
    void requestUpdate( void );

    //* update
    void update( const Job&, const QString& = QString() );

    //* update status
    void updateStatus( const Job& );

    //* update status
    void updateStatus( QString, const JobSet& );

    //* update status
    void updateStatus( const Job&, const JobSet& );

    //* update jdl
    void updateJdl( const Job&, const QString& = QString() );

    //* update output files
    void updateFiles( const Job&, const QString&, const FileRecord::List& );

    //* command completed
    void commandCompleted( const AlienCommand& );

    protected Q_SLOTS:

    //* open or save
    void _openOrSave( OpenFileDialog::Commands = OpenFileDialog::Any );

    //* open
    void _open( void );

    //* save as
    void _saveAs( File = File() );

    //* delete output directory
    void _deleteOutputDirectory( void );

    //* view output directory
    void _viewOutputDirectory( void );

    //* resubmit job
    void _resubmit( void );

    //* update actions
    void _updateActions( void );

    //* open link
    void _openLink( QString );

    private:

    //* request output files
    void _requestFiles( const QString& );

    //* update tabs visibility
    void _updateTabsVisibility( void );

    //* recycle path, from job information
    QString _recyclePath( void ) const;

    private:

    //* install actions
    void _installActions( void );

    //* job
    Job job_;

    //* tab widget
    QTabWidget* tabWidget_ = nullptr;

    //* id
    GridLayoutItem* idLabel_ = nullptr;

    //* parent
    GridLayoutItem* parentLabel_ = nullptr;

    //* user
    GridLayoutItem* userLabel_ = nullptr;

    //* command
    GridLayoutItem* commandLabel_ = nullptr;

    //* state
    GridLayoutItem* statusLabel_ = nullptr;

    //* submission time
    GridLayoutItem* submissionTimeLabel_ = nullptr;

    //* start time
    GridLayoutItem* startTimeLabel_ = nullptr;

    //* completion time
    GridLayoutItem* completionTimeLabel_ = nullptr;

    //* recycle path
    GridLayoutItem* outputPathLabel_ = nullptr;

    //* tools panel
    Private::ToolsPanel* tools_ = nullptr;

    //* trace
    TextEditor* traceEditor_ = nullptr;

    //* jdl
    TextEditor* jdlEditor_ = nullptr;

    //* syntax highlighter
    JdlSyntaxHighlighter* jdlSyntaxHighlighter_ = nullptr;

    //* files container
    QWidget* filesContainer_ = nullptr;

    //* recycle path
    GridLayoutItem* pathLabel_ = nullptr;

    //* busy widget
    BusyWidget* busyWidget_ = nullptr;

    //* job output files
    TreeView* list_ = nullptr;

    //*@name actions
    //@{

    //* open action
    QAction* openAction_ = nullptr;

    //* save as action
    QAction* saveAsAction_ = nullptr;

    //@}

    //* model
    class Model: public ListModel<FileRecord>
    {

        public:

        //* column type enumeration
        enum ColumnType
        {
            Permissions,
            User,
            Group,
            Size,
            Creation,
            Filename,
            nColumns
        };

        //*@name methods reimplemented from base class
        //@{

        // return data for a given index
        virtual QVariant data( const QModelIndex&, int ) const;

        //* header data
        virtual QVariant headerData( int, Qt::Orientation, int = Qt::DisplayRole ) const;

        //* number of columns for a given index
        virtual int columnCount( const QModelIndex& = QModelIndex() ) const
        { return nColumns; }

        //@}

        protected:

        //* sort
        virtual void _sort( int column, Qt::SortOrder order = Qt::AscendingOrder );

        private:

        //* used to sort IconCaches
        class SortFTor: public ItemModel::SortFTor
        {

            public:

            //* constructor
            SortFTor( const int& type, Qt::SortOrder order = Qt::AscendingOrder ):
                ItemModel::SortFTor( type, order )
            {}

            //* prediction
            bool operator() ( FileRecord, FileRecord ) const;

        };

        //* column titles
        static const QString columnTitles_[ nColumns ];

    };

    //* model
    Model model_;

    //* default user
    QString defaultUser_;

    //* output directory
    QString outputDirectory_;

    //* command ids
    QSet<int> commands_;

};

#endif
