/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "JobManager.h"

#include "Command.h"
#include "Debug.h"
#include "File.h"
#include "Job.h"
#include "InformationDialog.h"
#include "TreeView.h"
#include "Util.h"
#include "XmlOptions.h"

#include <QApplication>
#include <QHeaderView>
#include <QScrollBar>

//_______________________________________________
JobManager::JobManager( QWidget *parent ):
    QWidget( parent )
{
    Debug::Throw("JobManager::JobManager.\n");

    // layout
    setLayout( new QVBoxLayout( this ) );
    layout()->setSpacing(0);
    layout()->setMargin(0);

    // list
    TreeView::Container* container = new TreeView::Container( this );
    layout()->addWidget( container );

    treeView_ = &container->treeView();
    treeView_->setMouseTracking( true );
    treeView_->setModel( &model_ );
    treeView_->setSortingEnabled( true );
    treeView_->setRootIsDecorated( true );
    treeView_->setSelectionMode( TreeView::ContiguousSelection );
    treeView_->setOptionName( "JOBLIST" );

    treeView_->lockColumnVisibility( JobModel::Empty );
    treeView_->lockColumnVisibility( Job::User );
    treeView_->setColumnHidden( JobModel::Empty, false );
    treeView_->setColumnHidden( Job::User, true );

    connect( treeView_, SIGNAL(expanded(QModelIndex)), treeView_, SLOT(resizeColumns()) );
    connect( treeView_, SIGNAL(expanded(QModelIndex)), SLOT(_requestChildrenList(QModelIndex)) );
    connect( treeView_, SIGNAL(collapsed(QModelIndex)), treeView_, SLOT(resizeColumns()) );

    connect( &model_, SIGNAL(layoutAboutToBeChanged()), SLOT(_lockRequests()) );
    connect( &model_, SIGNAL(layoutChanged()), SLOT(_unlockRequests()) );

}

//_______________________________________
bool JobManager::setUser( QString user )
{

    Debug::Throw() << "JobManager::setUser - " << user << endl;
    if( user.isEmpty() || user == user_ ) return false;
    user_ = user;
    model_.clear();
    return true;

}

//____________________________________________________________
JobSet JobManager::updateMasterList( JobSet jobs )
{
    Debug::Throw( "JobManager::updateMasterList.\n" );

    // update children statistics from model
    _updateTimeStamps( Job(), jobs );
    _updateChildrenStatistics( jobs );

    // keep track of jobs that are missing time stamps
    const JobSet incompleteJobs( _incompleteJobs( jobs ) );

    // keep tracks of jobs to be deleted from model
    JobModel::List deletedJobs;
    Job::IdSet deletedIds;

     // retrieve current set
    for( const auto& currentJob:model_.children() )
    {

        // skip children
        if( currentJob.hasParent() )
        {

            if( deletedIds.contains( currentJob.parentId() ) )
            {
                deletedJobs << currentJob;
                deletedIds.insert( currentJob.id() );
            }

        } else if( !jobs.contains( currentJob ) ) {

            deletedJobs << currentJob;
            deletedIds.insert( currentJob.id() );

        }

    }

    // busy
    emit busy( jobs.size() );

    // removed jobs that are not found
    model_.remove( deletedJobs );

    // assign to model
    model_.add( jobs.toList() );

    // resize columns
    treeView_->resizeColumns();

    // re-emit
    emit jobListAvailable( user(), JobSet().fromList( model_.children() ) );

    // iddle
    emit idle();
    emit messageAvailable( QString( tr( "Last update: %1" ) ).arg( TimeStamp::now().toString( TimeStamp::TimeLong ) ) );

    return incompleteJobs;
}

//____________________________________________________________
JobSet JobManager::updateChildrenList( const Job& master, JobSet jobs )
{

    Debug::Throw( "JobManager::updateChildrenList.\n" );

    // children statistics
    _updateTimeStamps( master, jobs, false );
    _updateChildrenStatistics( master, jobs );

    // keep track of jobs that are missing time stamps
    const JobSet incompleteJobs( _incompleteJobs( jobs ) );

    // assign to model
    model_.setToday( TimeStamp::now() );
    model_.set( master, jobs.toList() );

    // resize columns
    treeView_->resizeColumns();

    // re-emit
    emit jobListAvailable( user(), JobSet().fromList( model_.children() ), false );

    return incompleteJobs;

}

//____________________________________________________________
JobSet JobManager::updateJobList( JobSet jobs )
{
    Debug::Throw( "JobManager::updateJobList.\n" );

    // update timestamps
    _updateTimeStamps( Job(), jobs );

    // keep track of jobs that are missing time stamps
    const JobSet incompleteJobs( _incompleteJobs( jobs ) );

    // busy
    emit busy( jobs.size() );

    // check children (this updates master jobs statistics)
    jobs.checkChildren();

    // assign to model
    model_.setToday( TimeStamp::now() );
    model_.set( jobs.toList() );

    // re-emit
    emit jobListAvailable( user(), jobs );

    // iddle
    emit idle();

    emit messageAvailable( QString( tr( "Last update: %1" ) ).arg( TimeStamp::now().toString( TimeStamp::TimeLong ) ) );

    return incompleteJobs;
}

//________________________________________
JobSet JobManager::selectedJobs( bool check )
{
    Debug::Throw( "JobManager::selectedJobs.\n" );
    JobSet selected;

    // retrieve selected indexes in list
    QModelIndexList selectedIndexes( treeView_->selectionModel()->selectedIndexes() );
    for( const auto& index:selectedIndexes )
    {

        // check column
        if( !index.column() == 0 ) continue;
        if( !treeView_->isVisible( index ) ) continue;

        // get current job
        const Job& job( model_.get( index ) );

        selected.insert( job );

    }

    if( check && selected.empty() )
    {
        // popup warning when no job is selected (nor removed by "permission denied" )
        QString buffer( "No selected jobs." );
        InformationDialog( this, buffer ).exec();
    }

    return selected;

}

//________________________________________
JobSet JobManager::masterJobs( void )
{
    Debug::Throw( "JobManager::masterJobs.\n" );
    JobSet masterJobs;
    for( const auto& job:model_.children() )
    { if( job.isMasterJob() ) masterJobs.insert( job ); }

    return masterJobs;

}

//____________________________________________________________
void JobManager::processJobInformation( const Job& job )
{

    // check user
    if( job.get( Job::User ) != user_ ) return;
    Debug::Throw() << "JobManager::processJobInformation - job " << job.id() << endl;

    // get index from job
    QModelIndex index( model_.index( job ) );
    if( !index.isValid() ) return;

    // get job
    Job local( model_.get( index ) );
    if( job.has( Job::SubmissionTime ) ) local.assign( Job::SubmissionTime, job.get( Job::SubmissionTime ) );
    if( job.has( Job::StartTime ) ) local.assign( Job::StartTime, job.get( Job::StartTime ) );
    if( job.has( Job::EndTime ) ) local.assign( Job::EndTime, job.get( Job::EndTime ) );

    if( job.hasParent() && local.parentId() != job.parentId() )
    {


        local.setParentId( job.parentId() );

        // need to re-run "checkChildren"
        JobModel::List currentJobs( model_.children() );
        JobSet jobs( JobSet::fromList( currentJobs ) );
        jobs.remove( local );
        jobs.insert( local );
        jobs.checkChildren();

        // re-emit
        emit jobListAvailable( user(), jobs );

        // assign to model
        model_.setToday( TimeStamp::now() );
        model_.set( jobs.toList() );

    } else {

        // directly insert in list
        Debug::Throw( "JobManager::updateJob - inserting local.\n" );
        model_.add( local );

    }

}

//____________________________________________________________
void JobManager::processChildrenInformation( const Job& job, const JobSet& children )
{

    // check user
    if( job.get( Job::User ) != user_ ) return;

    // check job index
    QModelIndex index( model_.index( job ) );
    if( !index.isValid() ) return;

    Debug::Throw() << "JobManager::processChildrenInformation - job " << job.id() << endl;

    // get children and update time stamps
    bool modified( false );
    JobModel::List currentJobs( model_.children( index ) );
    for( auto& job:currentJobs )
    {
        JobSet::const_iterator childIter( children.find( job ) );
        if( childIter == children.end() ) continue;

        if( childIter->has( Job::SubmissionTime ) ) { modified = true; job.assign( Job::SubmissionTime, childIter->get( Job::SubmissionTime ) ); }
        if( childIter->has( Job::StartTime ) ) { modified = true; job.assign( Job::StartTime, childIter->get( Job::StartTime ) ); }
        if( childIter->has( Job::EndTime ) ) { modified = true; job.assign( Job::EndTime, childIter->get( Job::EndTime ) ); }
        if( childIter->has( Job::Resubmitted ) ) { modified = true; job.assign( Job::Resubmitted, childIter->get( Job::Resubmitted ) ); }
    }

    // reassign to model
    if( modified )
    {
        model_.setToday( TimeStamp::now() );
        model_.set( job, currentJobs );
    }

}

//____________________________________________________
void JobManager::_updateTimeStamps( const Job& master, JobSet& jobs, bool emitProgress )
{

    Debug::Throw( "JobManager::_updateTimeStamps.\n" );

    // retrieve current set
    const JobModel::List currentJobs( master.isValid() ? model_.children( model_.index( master ) ):model_.children() );

    if( currentJobs.empty() ) return;
    if( emitProgress ) emit busy( currentJobs.size() );
    int index(0);
    for( const auto& currentJob:currentJobs )
    {

        // progress
        if( emitProgress ) emit progressAvailable( index );
        ++index;

        // skip if timestamps are not set
        if( !(
            currentJob.has( Job::SubmissionTime ) ||
            currentJob.has( Job::StartTime ) ||
            currentJob.has( Job::EndTime ) ||
            currentJob.has( Job::Resubmitted ) ||
            currentJob.hasParent() ) )
        { continue; }

        // find in current job set
        JobSet::const_iterator iter( jobs.find( currentJob ) );
        if( iter != jobs.end() )
        {
            const Job& job( *iter );

            // copy parent if needed
            if( currentJob.hasParent() && currentJob.parentId() != iter->parentId() )
            { const_cast<Job*>(&job)->setParentId( currentJob.parentId() ); }

            // copy timestamps if needed
            if( currentJob.has( Job::SubmissionTime ) && !iter->has( Job::SubmissionTime ) ) { const_cast<Job*>(&job)->assign( Job::SubmissionTime, currentJob.get( Job::SubmissionTime ) ); }
            if( currentJob.has( Job::StartTime ) && !iter->has( Job::StartTime ) ) { const_cast<Job*>(&job)->assign( Job::StartTime, currentJob.get( Job::StartTime ) ); }
            if( currentJob.has( Job::EndTime ) && !iter->has( Job::EndTime ) ) { const_cast<Job*>(&job)->assign( Job::EndTime, currentJob.get( Job::EndTime ) ); }
            if( currentJob.has( Job::Resubmitted ) ) { const_cast<Job*>(&job)->assign( Job::Resubmitted, currentJob.get( Job::Resubmitted ) ); }

        }

    }

    // perform second loop to fill blanks
    const QString timestamp( TimeStamp::now().toString( "MMM d hh:mm" ) );

    for( const auto& job:jobs )
    {

        // assign run time if missing
        if( ( job.isRunning() || job.isCompleted() || job.isFailed() ) && !job.has( Job::StartTime ) )
        { const_cast<Job&>(job).assign( Job::StartTime, timestamp ); }

        // assign end time if missing
        if( ( job.isCompleted() || job.isFailed() ) && !job.has( Job::EndTime ) )
        { const_cast<Job&>(job).assign( Job::EndTime, timestamp ); }

    }

    if( emitProgress ) emit idle();

}

//____________________________________________________
void JobManager::_updateChildrenStatistics( JobSet& jobs )
{

    Debug::Throw( "JobManager::_updateChildrenStatistics.\n" );

    // retrieve current set
    for( const auto& currentJob:model_.children() )
    {

        // skip if timestamps are not set
        if( !currentJob.isMasterJob() )
        { continue; }

        // find in current job set
        JobSet::const_iterator iter( jobs.find( currentJob ) );
        if( iter != jobs.end() )
        {
            Job job( *iter );

            // copy children statistics
            job.setChildrenStatistics( currentJob.childrenStatistics() );

            // copy job command
            if( currentJob.has( Job::Command ) )
            { job.assign( Job::Command, currentJob.get( Job::Command ) ); }

            jobs.remove( *iter );
            jobs.insert( job );

        }

    }

}

//____________________________________________________
void JobManager::_updateChildrenStatistics( Job master, JobSet jobs )
{

    Debug::Throw( "JobManager::_updateChildrenStatistics.\n" );

    // store command
    const QString command( !jobs.empty() && jobs.begin()->has( Job::Command ) ? jobs.begin()->get( Job::Command ):QString() );

    // insert master in joblist, check children and update
    jobs.insert( master );
    jobs.checkChildren();
    master = *jobs.find( master );

    // retrive model job
    Job modelJob( model_.get( model_.index( master ) ) );
    modelJob.setChildrenStatistics( master.childrenStatistics() );

    // update command if needed
    if( !command.isEmpty() )
    { modelJob.assign( Job::Command, command ); }

    // re-insert in model
    model_.add( modelJob );

}

//_____________________________________________________________
JobSet JobManager::_incompleteJobs( const JobSet& jobs ) const
{

    JobSet out;
    for( const auto& job:jobs )
    {

        // skip killed jobs, since they will dissapear
        if( job.isKilled() ) continue;

        // check for missing submission time
        if( !job.has( Job::SubmissionTime ) ) out.insert( job );

        // check for missing run time
        else if( ( ( job.isRunning() && !job.isInserting() ) || job.isCompleted() || job.isFailed() ) && !job.has( Job::StartTime ) ) out.insert( job );

        // check for missing end time
        else if( ( job.isCompleted() || job.isFailed() ) && !job.has( Job::EndTime ) ) out.insert( job );

    }

    return out;

}

//________________________________________
void JobManager::_requestChildrenList( const QModelIndex& index )
{

    // check lock
    if( requestsLocked_ ) return;
    Debug::Throw( "JobManager::_requestChildrenList.\n" );

    // check index
    if( !index.isValid() ) return;

    // retrieve job, check if master and request children list
    const Job job( model_.get( index ) );
    if( job.isMasterJob() ) emit childrenListRequested( job );

}
