/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "MainWindow.h"

#include "AlienCommandDialog.h"
#include "AlienInterface.h"
#include "Application.h"
#include "CommandsFailedDialog.h"
#include "ContextMenu.h"
#include "CustomToolBar.h"
#include "File.h"
#include "IconNames.h"
#include "IconEngine.h"
#include "InformationDialog.h"
#include "JobCollection.h"
#include "JobInformationDialog.h"
#include "JobListDialog.h"
#include "JobManager.h"
#include "JobManagerFrame.h"
#include "JobSet.h"
#include "JobStatistics.h"
#include "LogDialog.h"
#include "LogWidget.h"
#include "Menu.h"
#include "NewFrameDialog.h"
#include "ProgressStatusBar.h"
#include "QuestionDialog.h"
#include "QtUtil.h"
#include "SummaryFrame.h"
#include "Singleton.h"
#include "TreeView.h"
#include "UserSelectionFrame.h"
#include "UserSet.h"
#include "Util.h"
#include "WarningDialog.h"
#include "XmlDef.h"
#include "XmlOptions.h"

//___________________________________________
class TabWidget: public QTabWidget
{

    public:

    //! constructor
    TabWidget( QWidget* parent ):
        QTabWidget( parent )
    {}

    protected:

    //! tab insertion
    virtual void tabInserted( int index )
    {
        QTabWidget::tabInserted( index );
        int count( QTabWidget::count() );
        if( tabBar() ) tabBar()->setVisible( count > 1 );
    }


    //! tab insertion
    virtual void tabRemoved( int index )
    {
        QTabWidget::tabRemoved( index );
        int count( QTabWidget::count() );
        if( tabBar() ) tabBar()->setVisible( count > 1 );
    }

};

//___________________________________________
MainWindow::MainWindow( QWidget* parent ):
    BaseMainWindow( parent )
{
    Debug::Throw( "MainWindow::MainWindow.\n" );

    setOptionName( "MAIN_WINDOW" );

    // interface
    interface_ = new AlienInterface( this );
    interface_->findCertificates();

    // install actions
    _installActions();
    Application& application( *static_cast<Application*>(Singleton::get().application()) );
    addAction( &application.closeAction() );

    // statusBar
    setStatusBar( statusBar_ = new ProgressStatusBar( this ) );
    statusBar_->addClock();

    // user toolbar
    userToolBar_ = new CustomToolBar( "User Toolbar", this, "USER_TOOLBAR" );
    userToolBar_->addWidget( userSelectionFrame_ = new UserSelectionFrame( userToolBar_ ) );
    userSelectionFrame_->comboBox().setEditable( false );

    // add default users to user_selection frame
    userSelectionFrame_->updateUsers( UserSet() );

    // main toolbar
    toolbar_ =  new CustomToolBar( "Main Toolbar", this, "TOOLBAR" );
    toolbar_->addAction( addFrameAction_ );
    toolbar_->addSeparator();
    toolbar_->addAction( resubmitAction_ );
    toolbar_->addAction( killAction_ );
    toolbar_->addAction( traceAction_ );
    toolbar_->addSeparator();
    toolbar_->addAction( updateAction_ );
    toolbar_->addAction( pauseAction_ );

    toolbar_->addSeparator();
    toolbar_->addAction( clearAction_ );
    toolbar_->addAction( abortAction_ );

    // main widget and layout
    QWidget* main( new QWidget( this ) );
    main->setLayout( new QVBoxLayout() );
    main->layout()->setMargin(0);
    main->layout()->setSpacing(0);
    setCentralWidget( main );

    // tab widget
    main->layout()->addWidget( tabWidget_ = new TabWidget( main ) );
    tabWidget_->setMovable( true );

    connect( tabWidget_, SIGNAL(currentChanged(int)), SLOT(_activeFrameChanged(int)) );

    // create removeTab button
    #if QT_VERSION < 0x040500
    {
        QToolButton* button = new QToolButton();
        button->setDefaultAction( &removeFrameAction() );
        button->setAutoRaise( true );
        tabWidget_->setCornerWidget( button, Qt::TopRightCorner );
    }
    #else
    tabWidget_->setTabsClosable( true );
    connect( tabWidget_, SIGNAL(tabCloseRequested(int)), SLOT(_removeFrame(int)));
    #endif

    // menu bar
    setMenuBar( new Menu( this ) );
    connect( &userSelectionFrame(), SIGNAL(userChanged(QString)), SLOT(_updateUser(QString)) );

    // log dialog
    logDialog_ = new LogDialog();
    connect( logDialog_, SIGNAL(dialogHidden()), SLOT(_logDialogHidden()) );

    // interface connections
    connect( this, SIGNAL(commandRequested(AlienCommand)), interface_, SLOT(addCommand(AlienCommand)) );
    connect( this, SIGNAL(commandsRequested(const AlienCommand::List&)), interface_, SLOT(addCommands(const AlienCommand::List&)) );
    connect( abortAction_, SIGNAL(triggered()), interface_, SLOT(abortAll()) );
    connect( this, SIGNAL(activeUsersChanged(UserSet)), interface_, SLOT(updateActiveUsers(UserSet)) );

    connect( interface_, SIGNAL(defaultUserAvailable(QString)), this, SLOT(_updateDefaultUser(QString)) );
    connect( interface_, SIGNAL(userListAvailable(UserSet)), this, SLOT(_processUserList(UserSet)) );
    connect( interface_, SIGNAL(masterListAvailable(QString,JobSet)), this, SLOT(_processMasterList(QString,JobSet)) );
    connect( interface_, SIGNAL(childrenListAvailable(Job,JobSet)), this, SLOT(_processChildrenList(Job,JobSet)) );
    connect( interface_, SIGNAL(messageAvailable(QString)), SLOT(_updateStatusBar(QString)));

    connect( interface_, SIGNAL(commandAdded(AlienCommand)), &model_, SLOT(add(AlienCommand)) );
    connect( interface_, SIGNAL(commandsAdded(const AlienCommand::List&)), &model_, SLOT(add(const AlienCommand::List&)) );
    connect( interface_, SIGNAL(commandCompleted(AlienCommand)), this, SLOT(_commandCompleted(AlienCommand)) );
    connect( interface_, SIGNAL(commandsFailed(const AlienCommand::List&)), this, SLOT(_commandsFailed(const AlienCommand::List&)) );
    connect( interface_, SIGNAL(commandCompleted(AlienCommand)), &model_, SLOT(remove(AlienCommand)) );

    // connection state
    connect( interface_, SIGNAL(connectionStateChanged(bool)), this, SLOT(_connectionStateChanged(bool)) );
    connect( interface_, SIGNAL(connectionLost(QString)), this, SLOT(_lostConnection(QString)) );

    // busy state
    connect( interface_, SIGNAL(busyStateChanged(bool)), abortAction_, SLOT(setEnabled(bool)) );
    connect( interface_, SIGNAL(busy()), this, SLOT(_busy()) );
    connect( interface_, SIGNAL(ready()), this, SLOT(_idle()) );

    // configuration update
    connect( &application, SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    connect( &application, SIGNAL(saveConfiguration()), SLOT(_saveConfiguration()) );
    connect( qApp, SIGNAL(aboutToQuit()), SLOT(_saveConfiguration()) );
    _updateConfiguration();

    // busy widget
    busyWidget_ = new BusyWidget( main, BusyWidget::BottomRight );
    busyWidget_->hide();

    // initialize
    _checkProxy();
    _requestDefaultUser();
    _requestUserList();
    _requestJobList();

}

//_______________________________________
void MainWindow::installFrames( void )
{

    Debug::Throw( "MainWindow::installFrames.\n" );
    _updateStatusBar( tr( "Loading saved jobs" ) );

    JobCollection collection;
    collection.read( File( XmlOptions::get().raw( "RC_FILE" ) ) );

    JobManagerFrame* firstFrame(0);
    const JobCollection::List& jobSets( collection.jobSets() );
    if( jobSets.empty() )
    {

        // create new Frame
        JobManagerFrame& frame( _newFrame() );
        if( !firstFrame ) firstFrame = &frame;
        frame.setUser( _hasDefaultUser() ? defaultUser_ : UserSet::defaultUser() );
        tabWidget_->addTab( &frame, frame.jobManager().user() );

    } else for( const auto& jobSet:jobSets ) {

        // create new Frame
        JobManagerFrame& frame( _newFrame() );
        if( jobSet.isActive() || !firstFrame ) firstFrame = &frame;

        // assign user
        if( UserSet::isDefaultUser( jobSet.owner() ) && _hasDefaultUser() ) frame.setUser( defaultUser_ );
        else frame.setUser( jobSet.owner() );

        // assign jobs
        frame.jobManager().updateJobList( jobSet );

        // add to tab widget
        tabWidget_->addTab( &frame, frame.jobManager().user() );

    }

    Q_CHECK_PTR( firstFrame );
    _setActiveFrame( *firstFrame );
    firstFrame->jobManager().list().setFocus();
    tabWidget_->setCurrentWidget( firstFrame );
    _updateStatusBar( QString() );

    // notify change of active users
    emit activeUsersChanged( _users() );

}

//_______________________________________________________________________
void MainWindow::_checkProxy( void )
{
    Debug::Throw( "MainWindow::_checkProxy.\n" );
    emit commandsRequested(
        AlienCommand::List()
        << AlienCommand::CheckProxy
        << AlienCommand::FlushErrors );

}

//_______________________________________________________________________
void MainWindow::_connect( void )
{
    Debug::Throw( "MainWindow::_connect.\n" );
    emit commandRequested( AlienCommand( AlienCommand::Connect ) );
}

//____________________________________________________________________________________________
void MainWindow::_connectionStateChanged( bool connected )
{
    Debug::Throw() << "MainWindow::_connectionStateChanged" << endl;

    if( connected && !wasConnected_ ) wasConnected_ = true;
    if( connected && !pauseAction_->isChecked() )
    {
        if( !timer_.isActive() ) timer_.start( refreshRate_, this );
        _requestDefaultUser();
        _requestUserList();
    }

    _updateActions();
    _updateClearAction();

}

//____________________________________________________________________________________________
void MainWindow::_lostConnection( const QString& details )
{
    Debug::Throw( "MainWindow::_lostConnection.\n" );
    QString message( wasConnected_ ?
        tr( "Lost connection to remote file system.\nTry to reconnect ?" ):
        tr( "No connection to remote file system. Try connect ?" ) );
    QuestionDialog dialog( this, message );
    dialog.setOptionName( "CONNECTION_LOST_DIALOG" );
    dialog.setDetails( details );
    dialog.okButton().setText( tr( "Connect" ) );
    dialog.okButton().setIcon( IconEngine::get( IconNames::Connect ) );
    if( dialog.exec() ) _connect();

    return;
}

//___________________________________________________________
void MainWindow::_closeJobInformationDialogs( void )
{

    Debug::Throw( "MainWindow::_closeJobInformationDialogs.\n" );

    // need local copy of dialog map because the latter might get modified while the dialogs are closed
    for( const auto& dialog:Base::KeySet<JobInformationDialog>( this ) )
    { dialog->close(); }

}

//____________________________________________________________
void MainWindow::_requestDefaultUser( void )
{
    Debug::Throw( "MainWindow::_requestDefaultUser.\n" );
    emit commandRequested( AlienCommand( AlienCommand::DefaultUser ) );
}

//____________________________________________________________
void MainWindow::_requestUserList( void )
{
    Debug::Throw( "MainWindow::_requestUserList.\n" );
    emit commandRequested( AlienCommand( AlienCommand::UserList ) );
}

//____________________________________________________________
void MainWindow::_requestJobList( void )
{
    Debug::Throw( "MainWindow::_requestJobList.\n" );

    // store managers that must be updated
    QStringList users;

    // start with active manager
    if( _hasActiveFrame() ) users << _activeFrame().jobManager().user();

    // add managers for which summary is detached
    for( const auto& frame:Base::KeySet<JobManagerFrame>( this ) )
    {

        // check frame
        if( !frame->summaryFrame().isDetached() ) continue;
        if( _isActiveFrame( *frame ) ) continue;

        // append
        users << frame->jobManager().user();

    }

    // todo also implement other frames
    _requestJobList( users );

}


//____________________________________________________________
void MainWindow::_updateDefaultUser( QString user )
{
    Debug::Throw() << "MainWindow::_updateDefaultUser - user: " << user << endl;

    // save
    defaultUser_ = user;

    // update frame names
    for( const auto& frame:Base::KeySet<JobManagerFrame>( this ) )
    {
        if( UserSet::isDefaultUser( frame->jobManager().user() ) )
        { frame->setUser( user ); }
    }

    // update default user
    for( const auto& dialog:Base::KeySet<JobInformationDialog>( this ) )
    { dialog->setDefaultUser( user ); }

    // update user selection frame
    if( UserSet::isDefaultUser( userSelectionFrame_->user() ) )
    { userSelectionFrame_->setUser( defaultUser_ ); }

    // request job list
    _requestJobList();

}

//____________________________________________________________
void MainWindow::_processUserList( UserSet users )
{

    Debug::Throw( "MainWindow::_processUserList.\n" );

    if( _hasDefaultUser() )
    {

        // remove default string
        users.setNeedsDefault( false );

        // make sure defaulut user is included
        if( !users.contains( defaultUser_ ) ) users.append( defaultUser_ );
        Debug::Throw() << "MainWindow::_processUserList - users: " << users.join("\n") << endl;
    }

    userSelectionFrame_->updateUsers( users );

    // install frames
    if( Base::KeySet<JobManagerFrame>(this).empty() ) installFrames();

    // need to reemit the signal so that it gets catched
    // by possible newFrame dialogs
    emit userListAvailable( users );

}

//____________________________________________________________
void MainWindow::_processMasterList( QString user, const JobSet& jobs )
{

    Debug::Throw() << "MainWindow::_processMasterList - user: " << user << endl;
    JobManagerFrame* frame( _findFrame( user ) );
    if( !frame ) return;
    const bool isActiveFrame( _isActiveFrame( *frame ) );

    // insert in frame
    JobSet out( frame->jobManager().updateMasterList( jobs ) );

    // list of pending commands
    AlienCommand::List commands;

    // request all job children
    for( const auto& job:jobs )
    {
        const bool isExpanded( frame->jobManager().list().isExpanded( frame->jobManager().model().index( job ) ) );
        const int priority( (isActiveFrame && isExpanded) ? 0:1 );

        // check status to decide whether to submit children list request
        if( job.mightHaveChildren() )
        {
            AlienCommand command( AlienCommand::ChildrenList );
            command
                << user
                << QString::number(job.parentId())
                << QString::number(job.id());

            command.setPriority(priority);
            commands << command;
        }

    }

    // request information for jobs that need it
    if( isActiveFrame )
    {
        for( const auto& job:out )
        {
            const bool isExpanded( frame->jobManager().list().isExpanded( frame->jobManager().model().index( job ) ) );
            const int priority( (isActiveFrame && isExpanded) ? 0:1 );

            AlienCommand command( AlienCommand::Trace );
            command
                << user
                << QString::number(job.parentId() )
                << QString::number(job.id() );

            command.setPriority(priority);
            commands << command;

        }

    }

    // submit pending commands
    if( !commands.empty() ) emit commandsRequested( commands );

    // update statusbar if active frame summary is hidden
    if( _activeFrame().summaryFrame().isHidden() || _activeFrame().summaryFrame().isDetached() )
    { statusBar_->label().setTextAndUpdate( _activeFrame().summaryFrame().summaryString() ); }

    // update actions
    _updateActions();

    return;
}

//____________________________________________________________
void MainWindow::_processChildrenList( Job master, const JobSet& jobs )
{

    const QString user( master.get( Job::User ) );
    JobManagerFrame* frame( _findFrame( user ) );
    if( !frame ) return;

    const JobSet out( frame->jobManager().updateChildrenList( master, jobs ) );

    // resubmit failed jobs if any
    if( _canResubmit( master ) )
    {

        // get real master job from model
        JobModel& model( frame->jobManager().model() );
        Job master( model.get( model.index( master ) ) );

        // get resubmission commands
        const AlienCommand::List commands( _masterJobResubmissionCommands( master ) );
        if( !commands.isEmpty() )
        {
            // increment resubmission count and update model
            master.incrementResubmission();
            model.add( master );

            // submit commands
            emit commandsRequested( commands );
        }

    }

    if( _isActiveFrame( *frame ) )
    {

        if( !out.isEmpty() )
        {

            // submit job trace command to update incomplete jobs
            // with the correct priority
            const bool isExpanded( frame->jobManager().list().isExpanded( frame->jobManager().model().index( master ) ) );
            const int priority( isExpanded ? 0:1 );
            AlienCommand command( AlienCommand::Trace );
            command
                << user
                << QString::number(master.parentId() )
                << QString::number(master.id() );

            command.setPriority(priority);
            emit commandRequested( command );

        }

        // update clear action
        _updateClearAction();

    }

    // update statusbar if active frame summary is hidden
    if( _activeFrame().summaryFrame().isHidden() || _activeFrame().summaryFrame().isDetached() )
    { statusBar_->label().setTextAndUpdate( _activeFrame().summaryFrame().summaryString() ); }

    // update actions
    _updateActions();

    return;
}

//____________________________________________________________
void MainWindow::_commandCompleted( const AlienCommand& command )
{
    switch( command.type() )
    {

        // trigger job list update for kill and resubmit commands
        case AlienCommand::Kill:
        case AlienCommand::KillMaster:
        case AlienCommand::Resubmit:
        case AlienCommand::ResubmitMaster:
        {

            // get user
            const QString user( command.user() );

            // decide whether JobList update is needed
            bool updateNeeded( false );
            if( _hasActiveFrame() && _activeFrame().jobManager().user() == user ) updateNeeded = true;
            else {

                // loop over associated JobManagerFrames, try find one with matching user which is detached.
                // trigger update if found.
                for( const auto& frame:Base::KeySet<JobManagerFrame>( this ) )
                {

                    // check frame
                    if( frame->summaryFrame().isDetached() && frame->jobManager().user() == user )
                    {
                        updateNeeded = true;
                        break;
                    }

                }

            }

            if( updateNeeded )
            {
                // request new job list an clear queue
                _requestJobList( user );
                _requestClearQueue( user );
            }

        }
        break;

        case AlienCommand::ClearQueue:
        {
            const QString user( command.user() );
            if( JobManagerFrame* frame = _findFrame( user ) )
            { interface_->clearQueue( user, Job::IdSet( frame->jobManager().model().children() ) ); }
        }
        break;

        // do nothing for all other commands
        default: break;
    }

}

//____________________________________________________________________________________________
void MainWindow::_commandsFailed( const AlienCommand::List& commands )
{

    if( commands.isEmpty() ) return;
    Debug::Throw() << "MainWindow::_commandsFailed - " << commands.front().name() << endl;

    if( commands.size() == 1 && commands.front().type() == AlienCommand::CheckProxy )
    {

        WarningDialog dialog( 0, commands.front().errorString() );
        dialog.setDetails( commands.front().errorDetails() );
        dialog.setWindowTitle( tr( "Error - Gridwatch" ) );
        dialog.exec();
        interface_->clearFailedCommands();

    } else {

        CommandsFailedDialog dialog( 0, commands );
        dialog.setWindowTitle( tr( "Error - Gridwatch" ) );
        if( !dialog.exec() ) interface_->clearFailedCommands();

    }

    return;

}

//_______________________________________
void MainWindow::_updateActions( void )
{
    Debug::Throw( "MainWindow::_updateActions.\n" );


    defaultUserAction_->setEnabled( interface_->isConnected() && !pauseAction_->isChecked() );
    userListAction_->setEnabled( interface_->isConnected() && !pauseAction_->isChecked() );

    if( !interface_->isConnected() || pauseAction_->isChecked() )
    {

        killAction_->setEnabled( false );
        resubmitAction_->setEnabled( false );
        traceAction_->setEnabled( false );

    } else {

        // retrace action
        traceAction_->setEnabled( _hasActiveFrame() && !_activeFrame().jobManager().selectedJobs().isEmpty() );

        // resubmit and kill actions
        if( !_activeFrameIsDefault() )
        {

            resubmitAction_->setEnabled( false );
            killAction_->setEnabled( false );

        } else {

            // get list of selected jobs
            const JobSet selectedJobs( _activeFrame().jobManager().selectedJobs() );
            killAction_->setEnabled( !selectedJobs.empty() );

            bool resubmitEnabled( false );
            for( const auto& job:selectedJobs )
            {

                // trigger on either failed jobs, or master jobs with failed children
                if( job.hasParent() ||
                    ( job.isMasterJob() && job.childrenStatistics().get( JobStatistics::Failed ) > 0 ) ||
                    ( job.isMasterJob() && job.isFailed() ) )
                {
                    resubmitEnabled = true;
                    break;
                }

            }

            // enable action
            resubmitAction_->setEnabled( resubmitEnabled );

        }

    }

    return;

}

//___________________________________________________________
void MainWindow::_updateClearAction( void )
{
    Debug::Throw( "MainWindow::_updateClearAction.\n" );
    if( !interface_->isConnected() ) clearAction_->setEnabled( false );
    else if( !_activeFrameIsDefault() )
    {

        Debug::Throw( "MainWindow::_updateClearAction - invalid user\n" );
        clearAction_->setEnabled( false );
        return;

    } else {

        const JobModel::List jobs( _activeFrame().jobManager().model().children() );
        const bool hasCompletedJobs( std::find_if(  jobs.begin(), jobs.end(), Job::CompletedFTor() ) != jobs.end() );
        Debug::Throw() << "MainWindow::_updateClearAction - completed jobs:" << hasCompletedJobs << endl;

        clearAction_->setEnabled( hasCompletedJobs );

    }

}

//___________________________________________________________
void MainWindow::_updateUser( QString user )
{

    Debug::Throw() << "MainWindow::_updateUser - " << user << endl;
    if( !_hasActiveFrame() ) return;

    // automatically replace user name if default
    if( UserSet::isDefaultUser( user ) && _hasDefaultUser() ) user = defaultUser_;

    // check if there is a matching frame
    if( JobManagerFrame* frame = _findFrame( user ) )
    {

        tabWidget_->setCurrentWidget( frame );
        _setActiveFrame( *frame );

    } else {

        tabWidget_->setTabText( tabWidget_->currentIndex(), user );
        _activeFrame().setUser( user );

    }

    // notify change of active users
    emit activeUsersChanged( _users() );

    // request job list
    _requestJobList();

}

//___________________________________________________________
void MainWindow::_updateStatusBar( QString value )
{

    Debug::Throw( "MainWindow::_updateStatusBar.\n" );
    if( !( _hasActiveFrame() && ( _activeFrame().summaryFrame().isHidden() || _activeFrame().summaryFrame().isDetached() ) ) )
    { statusBar_->label().setTextAndUpdate( value ); }

}

//_______________________________________
void MainWindow::_togglePause( bool value )
{
    Debug::Throw() << "MainWindow::_togglePause - " << value << endl;

    // propagate to interface
    interface_->togglePause( value );

    // restart timer
    if( interface_->isConnected() && !value && !timer_.isActive() )
    { timer_.start( refreshRate_, this ); }

}

//_______________________________________
void MainWindow::_toggleSummary( bool value )
{

    Debug::Throw( "MainWindow::_toggleSummary.\n" );
    XmlOptions::get().set<bool>( "SHOW_SUMMARY", value );
    if( _hasActiveFrame() )
    {
        _activeFrame().summaryFrame().setVisible( value );
        if( value ) statusBar_->label().clear();
        else statusBar_->label().setTextAndUpdate( _activeFrame().summaryFrame().summaryString() );
    }
}

//_______________________________________
void MainWindow::_activeFrameChanged( int index )
{

    Debug::Throw() << "MainWindow::_activeFrameChanged - index: " << index << endl;
    Q_ASSERT( index >= 0 );
    JobManagerFrame* frame = qobject_cast<JobManagerFrame*>( tabWidget_->widget( index ) );
    Q_CHECK_PTR( frame );

    // change frame activity and update
    _setActiveFrame( *frame );

    // update userSelectionFrame so that it displays the correct user name
    userSelectionFrame_->setUser( frame->jobManager().user() );

    // request new job list
    if( !delayTimer_.isActive() )
    { delayTimer_.start( 1000, this ); }

    // update title
    setWindowTitle( QString( "%1 - Gridwatch" ).arg( frame->jobManager().user() ) );

    // update statusbar if active frame summary is hidden
    if( _activeFrame().summaryFrame().isHidden() || _activeFrame().summaryFrame().isDetached() )
    { statusBar_->label().setTextAndUpdate( _activeFrame().summaryFrame().summaryString() ); }

}

//_______________________________________
void MainWindow::_addFrame( void )
{
    Debug::Throw( "MainWindow::_addFrame.\n" );

    // create dialog
    NewFrameDialog dialog( this );

    // update user selection
    dialog.userSelectionFrame().updateUsers( userSelectionFrame_->users() );
    dialog.userSelectionFrame().setUser( userSelectionFrame_->user() );

    // connect for auto-update
    connect(
        this,
        SIGNAL(userListAvailable(QStringList)),
        &dialog.userSelectionFrame(),
        SLOT(updateUsers(QStringList)));

    _requestUserList();

    // adjust focus and map dialog
    dialog.userSelectionFrame().comboBox().lineEdit()->setFocus();
    dialog.userSelectionFrame().comboBox().lineEdit()->selectAll();
    if( !dialog.centerOnWidget( window() ).exec() ) return;

    // get user
    QString user( dialog.userSelectionFrame().user() );
    if( UserSet::isDefaultUser( user ) && _hasDefaultUser() ) user = defaultUser_;

    // check whether it is already displayed or not
    JobManagerFrame *frame(nullptr);
    if( (frame = _findFrame( user ) ) )
    {

        tabWidget_->setCurrentWidget( frame );

    } else {

        // create new Frame
        frame = &_newFrame();
        frame->setUser( user );

        tabWidget_->addTab( frame, user );
        tabWidget_->setCurrentWidget( frame );

    }

    // notify change of active users
    emit activeUsersChanged( _users() );

    // request job list
    _requestJobList( user );

}

//_______________________________________
void MainWindow::_removeFrame( void )
{ return _removeFrame( tabWidget_->currentIndex() ); }

//_______________________________________
void MainWindow::_removeFrame( int index )
{

    Debug::Throw( "MainWindow::_removeFrame.\n" );

    // check counts
    unsigned int counts( Base::KeySet<JobManagerFrame>( this ).size() );
    Q_ASSERT( counts > 1 && _hasActiveFrame() );

    // check frame
    Q_ASSERT( index >= 0 );
    JobManagerFrame* frame = qobject_cast<JobManagerFrame*>( tabWidget_->widget( index ) );
    Q_CHECK_PTR( frame );

    const QString buffer = QString( tr( "Remove tab '%1' ?" ) ).arg( frame->jobManager().user() );
    if( !QuestionDialog( this, buffer ).centerOnWidget( window() ).exec() ) return;

    // change remove action enability
    if( counts == 2 )
    { removeFrameAction_->setEnabled( false ); }

    // change active frame
    const int currentIndex( tabWidget_->currentIndex() );
    if( index == currentIndex )
    { tabWidget_->setCurrentIndex( ( index > 0 ) ? index-1:index ); }

    // remove
    tabWidget_->removeTab( index );

    // delete old active frame
    frame->deleteLater();

    // notify change of active users
    emit activeUsersChanged( _users() );

}

//____________________________________________________________
void MainWindow::_requestChildrenList( const Job& job )
{

    Debug::Throw() << "MainWindow::_requestChildrenList -"
        << " user: " << job.get( Job::User )
        << " job: " << job.id()
        << endl;

    AlienCommand::List commands;

    {
        // create Trace command and append to list
        AlienCommand command( AlienCommand::Trace );
        command
            << job.get( Job::User )
            << QString::number(job.parentId() )
            << QString::number(job.id() );
        commands << command;
    }

    {
        // create ChildrenList command and append to list
        AlienCommand command( AlienCommand::ChildrenList );
        command
            << job.get( Job::User )
            << QString::number( job.parentId() )
            << QString::number( job.id() );
        commands << command;
    }

    // submit pending commands
    emit commandsRequested( commands );

}

//____________________________________________________________
void MainWindow::_requestJobInformation( void )
{
    Debug::Throw( "MainWindow::_requestJobInformation.\n" );
    if( !_hasActiveFrame() ) return;

    // get current Job
    QModelIndex index( _activeFrame().jobManager().list().selectionModel()->currentIndex() );
    if( !index.isValid() ) return;

    const Job job( _activeFrame().jobManager().model().get( index ) );

    // retrieve associated job dialogs
    for( const auto& dialog:Base::KeySet<JobInformationDialog>( this ) )
    {
        if( dialog->job() == job )
        {
            dialog->requestUpdate();
            dialog->uniconify();
            return;
        }
    }

    // create dialog
    JobInformationDialog* dialog = new JobInformationDialog( 0 );
    Base::Key::associate( this, dialog );

    dialog->setJob( job );
    dialog->setDefaultUser( defaultUser_ );

    // setup connection
    connect( interface_, SIGNAL(jobInformationAvailable(Job,QString)), dialog, SLOT(update(Job,QString)) );
    connect( interface_, SIGNAL(masterListAvailable(QString,JobSet)), dialog, SLOT(updateStatus(QString,JobSet)) );
    connect( interface_, SIGNAL(childrenListAvailable(Job,JobSet)), dialog, SLOT(updateStatus(Job,JobSet)) );
    connect( interface_, SIGNAL(jdlAvailable(Job,QString)), dialog, SLOT(updateJdl(Job,QString)) );
    connect( interface_, SIGNAL(jobFilesAvailable(Job,QString,const FileRecord::List&)), dialog, SLOT(updateFiles(Job,QString,const FileRecord::List&)) );
    connect( interface_, SIGNAL(commandCompleted(AlienCommand)), dialog, SLOT(commandCompleted(AlienCommand)) );
    connect( dialog, SIGNAL(commandRequested(AlienCommand)), interface_, SLOT(addCommand(AlienCommand)) );
    connect( dialog, SIGNAL(commandsRequested(AlienCommand::List)), interface_, SLOT(addCommands(AlienCommand::List)) );
    connect( dialog, SIGNAL(scratchFileCreated(File)), this, SIGNAL(scratchFileCreated(File)) );

    // show dialog
    dialog->requestUpdate();
    dialog->centerOnPointer();
    dialog->show();
    dialog->toggleSticky( false );

}

//____________________________________________________________
void MainWindow::_kill( void )
{
    Debug::Throw( "MainWindow::kill.\n" );
    if( !_hasActiveFrame() ) return;

    // copy selected jobs
    const JobSet selectedJobs( _activeFrame().jobManager().selectedJobs() );

    // keep track of master jobs
    JobSet masterJobs;

    // store user
    const QString user( _activeFrame().jobManager().user() );

    // list of commands to be submitted
    AlienCommand::List commands;

    // first path to select master jobs
    for( const auto& job:selectedJobs )
    {

        // skip if not master
        if( !job.isMasterJob() ) continue;

        // create KillMaster command and append to list
        {
            AlienCommand command( AlienCommand::KillMaster );
            command
                << user
                << QString::number( job.parentId() )
                << QString::number( job.id() );
            commands << command;
        }

        // create Kill command and append to list
        {
            AlienCommand command( AlienCommand::Kill );
            command
                << user
                << QString::number( job.parentId() )
                << QString::number( job.id() );
            commands << command;
        }

        // insert in master list
        masterJobs.insert( job );

    }

    // second path to select subJobs
    for( const auto& job:selectedJobs )
    {

        // skip master jobs
        if( job.isMasterJob() ) continue;

        // skip if parent is already in master list
        if( masterJobs.contains( Job().assign<int>( Job::Id, job.parentId() ) ) ) continue;

        // create Kill command and append to list
        AlienCommand command( AlienCommand::Kill );
        command
            << user
            << QString::number( job.parentId() )
            << QString::number( job.id() );
        commands << command;

        masterJobs.insert( job );

    }

    // check commands
    if( commands.empty() ) return;

    // ask confirmation
    QuestionDialog* dialog;
    if( masterJobs.size() == 1 )
    {
        dialog = new QuestionDialog( this, QString( tr( "Kill job %1 ?" ) ).arg( masterJobs.begin()->id() ) );

    } else {

        dialog = new JobListDialog( this, masterJobs.toList() );
        dialog->setText( QString( tr( "Kill %1 jobs ?" ) ).arg( masterJobs.size() ) );

    }

    dialog->setWindowTitle( tr( "Kill Jobs - Gridwatch" ) );
    dialog->setOptionName( "KILL_JOBS_DIALOG" );
    dialog->okButton().setText( tr( "Kill" ) );
    dialog->okButton().setIcon( IconEngine::get( IconNames::Kill ) );
    if( dialog->exec() ) emit commandsRequested( commands );
    dialog->deleteLater();

}

//____________________________________________________________
void MainWindow::_resubmit( void )
{
    Debug::Throw( "MainWindow::_resubmit.\n" );
    if( !_hasActiveFrame() ) return;

    // check user
    if( !_activeFrameIsDefault() ) return;

    const JobSet selectedJobs( _activeFrame().jobManager().selectedJobs() );
    const QString user( _activeFrame().jobManager().user() );

    // store list of resubmitted master jobs
    Job::IdSet masterJobs;
    JobSet subJobs;
    JobModel::List failedJobs;

    //! list of commands to be emited
    AlienCommand::List commands;

    // first path to select master jobs
    for( const auto& job:selectedJobs )
    {

        // skip if not master
        if( !job.isMasterJob() || job.isFailed() )
        {
            subJobs.insert( job );
            continue;
        }

        // check children
        if( job.childrenStatistics().get( JobStatistics::Total ) <= 0 )
        {
            subJobs.insert( job );
            continue;
        }

        // get resubmission commands
        const AlienCommand::List jobCommands( _masterJobResubmissionCommands( job ) );
        if( !jobCommands.isEmpty() )
        {

            // merge command lists
            commands += jobCommands;

            // insert in master list
            masterJobs.insert( job.id() );

        }

    }

    // second path on subJobs
    // keep track of running jobs
    JobSet runningJobs;

    // also keep a different list for master jobs, used to submit ChildrenList command
    Job::IdSet masterJobs2( masterJobs );
    for( const auto& job:subJobs )
    {

        if( job.isFailed() )
        {

            // skip if parent is already in master list
            if( masterJobs.find( job.parentId() ) != masterJobs.end() ) continue;

            // create Resubmit command and append to list
            AlienCommand command( AlienCommand::Resubmit );
            command
                << user
                << QString::number( job.parentId() )
                << QString::number( job.id() );
            commands << command;

            failedJobs.append( job );

            // add parent in masterJobs list (used to re-request children
            if( job.hasParent() ) masterJobs2.insert( job.parentId() );

        } else runningJobs.insert( job );

    }

    // ask confirmation for running jobs
    if( !runningJobs.empty() )
    {
        QuestionDialog* dialog;
        if( runningJobs.size() == 1 )
        {

            dialog = new QuestionDialog( this, QString( tr( "Job %1 is not in a failed state.\nResubmit anyway ?" ) ).arg( runningJobs.begin()->id() ) );

        } else {

            dialog = new JobListDialog( this, runningJobs.toList() );
            dialog->setText( QString( tr( "Following %1 jobs are not in a failed state.\nResubmit anyway ?" ) ).arg( runningJobs.size() ) );

        }

        dialog->setWindowTitle( tr( "Resubmit Jobs - Gridwatch" ) );
        dialog->setOptionName( "RESUBMIT_JOBS_DIALOG" );
        dialog->okButton().setText( tr( "Resubmit" ) );
        dialog->okButton().setIcon( IconEngine::get( IconNames::Resubmit ) );
        if( dialog->exec() )
        {
            for( const auto& job:runningJobs )
            {

                // create Resubmit command and append to list
                AlienCommand command( AlienCommand::Resubmit );
                command
                    << user
                    << QString::number(job.parentId())
                    << QString::number(job.id() );
                commands << command;
                failedJobs.append( job );

                // also resubmit parsing of children
                if( job.hasParent() ) masterJobs2.insert( job.parentId() );

            }

        }

        dialog->deleteLater();

    }

    // reset timestamps of resubmitted jobs
    for( auto& job:failedJobs )
    {
        job.assign( Job::StartTime, QString() );
        job.assign( Job::EndTime, QString() );
    }

    activeFrame_->jobManager().model().add( failedJobs );

    // finally emit children and information requests
    for( const auto& id:masterJobs2 )
    {

        commands
            << ( AlienCommand( AlienCommand::ChildrenList )
            << user
            << "-1"
            << QString::number( id ) );

        commands
            << ( AlienCommand( AlienCommand::Trace )
            << user
            << "-1"
            << QString::number( id ) );

    }

    // submit pending commands
    if( !commands.empty() ) emit commandsRequested( commands );

}

//____________________________________________________________
void MainWindow::_clear( void )
{
    Debug::Throw( "MainWindow::clear.\n" );
    if( !_activeFrameIsDefault() ) return;

    // retrieve children from model and check wether at least one is failed
    const JobModel::List selectedJobs( _activeFrame().jobManager().model().children() );
    JobSet masterJobs;

    // store user
    const QString user( _activeFrame().jobManager().user() );

    // list of commands to be send
    AlienCommand::List commands;

    // first path to select master jobs
    for( const auto& job:selectedJobs )
    {

        // skip if not master
        if( !job.isMasterJob() ) continue;

        // retrieve children from model and check wether at least one is completed
        JobModel::List children( _activeFrame().jobManager().model().children( job ) );
        const int completedJobs( std::count_if( children.begin(), children.end(), Job::CompletedFTor() ) );
        if( completedJobs > 0 )
        {

            // insert in list to avoid duplicates in second path
            masterJobs.insert( job );

            // kill all DONE children
            AlienCommand command( AlienCommand::KillMaster );
            command
                << user
                << QString::number( job.parentId() )
                << QString::number( job.id() )
                << "DONE";
            commands << command;

        }

        // kill the job itself, if tagged completed and all subjobs are completed
        if( job.isCompleted() && completedJobs == children.size() )
        {
            AlienCommand command( AlienCommand::Kill );
            command
                << user
                << QString::number( job.parentId() )
                << QString::number( job.id() );
            commands << command;
        }

    }

    // second path to select other jobs
    for( const auto& job:selectedJobs )
    {

        // skip master jobs
        if( job.isMasterJob() ) continue;

        // skip if not failed
        if( !job.isCompleted() ) continue;

        // skip if parent is already in master list
        if( masterJobs.find( Job().assign<int>( Job::Id, job.parentId() ) ) != masterJobs.end() ) continue;

        // create Kill command and append to grid
        AlienCommand command( AlienCommand::Kill );
        command
            << user
            << QString::number( job.parentId() )
            << QString::number( job.id() );
        commands << command;
    }

    // submit pending commands
    if( !commands.empty() )
    {
        QuestionDialog dialog( this, tr( "Permanently remove all completed jobs from the list ?" ) );
        dialog.setWindowTitle( tr( "Remove Jobs - Gridwatch" ) );
        dialog.okButton().setText( tr( "Remove" ) );
        dialog.okButton().setIcon( IconEngine::get( IconNames::Clear ) );
        if( dialog.exec() )
        { emit commandsRequested( commands ); }

    }

}

//_______________________________________________
void MainWindow::_showCommandList( void )
{

    Debug::Throw( "MainWindow::_showCommandList.\n" );
    Base::KeySet<AlienCommandDialog> dialogs( this );

    if( dialogs.empty() )
    {
        AlienCommandDialog *dialog = new AlienCommandDialog(0);
        Base::Key::associate( dialog, this );

        dialog->setModel( &model_ );
        dialog->centerOnWidget( qApp->activeWindow() ).show();
        dialog->toggleSticky( false );
        dialog->raise();

        connect( this, SIGNAL(commandRequested(AlienCommand)), &dialog->list(), SLOT(resizeColumns()) );
        connect( interface_, SIGNAL(commandCompleted(AlienCommand)), &dialog->list(), SLOT(resizeColumns()) );
        connect( dialog, SIGNAL(destroyed()), SLOT(_commandListDestroyed()) );

        XmlOptions::get().set<bool>( "SHOW_COMMAND_LIST", true );

        return;

    } else {

        (*dialogs.begin())->uniconify();
        (*dialogs.begin())->raise();

    }

}

//_______________________________________________
void MainWindow::_showLog( void )
{
    Debug::Throw( "MainWindow::_showLog.\n" );
    logDialog_->show();
    logDialog_->toggleSticky( false );
    logDialog_->raise();
    XmlOptions::get().set<bool>( "SHOW_LOG", true );
}

//_______________________________________________
void MainWindow::_toggleLog( bool value )
{
    Debug::Throw( "MainWindow::_toggleLog.\n" );
    if( value )
    {

        connect( interface_, SIGNAL(logAvailable(QString,Format::TextFormatFlags)), &logDialog_->editor(), SLOT(append(QString,Format::TextFormatFlags)), Qt::UniqueConnection );

    } else {

        disconnect( interface_, SIGNAL(logAvailable(QString,Format::TextFormatFlags)), &logDialog_->editor(), SLOT(append(QString,Format::TextFormatFlags)) );
        logDialog_->hide();
        logDialog_->editor().clear();

    }

    logAction_->setEnabled( value );
    XmlOptions::get().set<bool>( "LOG_ENABLED", value );
}

//_______________________________________________
void MainWindow::_toggleHideKilledJobs( bool value )
{
    Debug::Throw( "MainWindow::_toggleHideKilledJobs.\n" );
    interface_->setHideKilledJobs( value );
    XmlOptions::get().set<bool>("HIDE_KILLED_JOBS", value );
}

//_______________________________________________
void MainWindow::_logDialogHidden( void )
{
    Debug::Throw( "MainWindow::_logDialogHidden.\n" );
    XmlOptions::get().set<bool>( "SHOW_LOG", false );
}

//_______________________________________________
void MainWindow::_commandListDestroyed( void )
{
    Debug::Throw( "MainWindow::_commandListDestroyed.\n" );
    Base::KeySet<AlienCommandDialog> dialogs( this );
    if( dialogs.empty() ) XmlOptions::get().set<bool>( "SHOW_COMMAND_LIST", false );
}

//____________________________________
void MainWindow::timerEvent( QTimerEvent *event )
{

    Debug::Throw( "MainWindow::timerEvent.\n" );
    if( event->timerId() == timer_.timerId() ||  event->timerId() == delayTimer_.timerId() )
    {
        if( interface_->isConnected() && !pauseAction_->isChecked() ) _requestJobList();
        if( delayTimer_.isActive() ) delayTimer_.stop();
    }

}

//____________________________________
void MainWindow::closeEvent( QCloseEvent *event )
{
    Debug::Throw( "MainWindow::closeEvent.\n" );
    _saveConfiguration();
    event->accept();
    qApp->quit();
}

//____________________________________________________________
JobManagerFrame* MainWindow::_findFrame( const QString& user ) const
{

    for( const auto& frame:Base::KeySet<JobManagerFrame>( this ) )
    {
        QString frameUser( frame->jobManager().user() );
        if( UserSet::isDefaultUser( frameUser ) ) frameUser = defaultUser_;
        if( frameUser == user ) return frame;
    }

    return 0;

}

//____________________________________________________________
bool MainWindow::_activeFrameIsDefault( void ) const
{ return _hasActiveFrame() && _hasDefaultUser() && _activeFrame().jobManager().user() == defaultUser_; }

//____________________________________________________________
void MainWindow::_requestJobList( QString user )
{

    Debug::Throw() << "MainWindow::_requestJobList - user: " << user << endl;
    if( UserSet::isDefaultUser( user ) ) user = defaultUser_;
    if( user.isEmpty() ) return;

    if( user == defaultUser_ ) emit commandRequested( AlienCommand( AlienCommand::DefaultMasterList ) << user );
    else emit commandRequested( AlienCommand( AlienCommand::MasterList ) << user );

}

//____________________________________________________________
void MainWindow::_requestJobList( QStringList users )
{
    Debug::Throw() << "MainWindow::_requestJobList - users: " << users.join( " " ) << endl;
    if( users.isEmpty() ) return;
    if( users.size() == 1 ) return _requestJobList( users.front() );
    else
    {
        AlienCommand command( AlienCommand::MasterList );
        for( auto user:users )
        {
            if( UserSet::isDefaultUser( user ) ) user = defaultUser_;
            if( user.isEmpty() ) continue;
            command << user;
        }

        // emit command
        if( !command.arguments().isEmpty() ) emit commandRequested( command );

    }
}

//____________________________________________________________
void MainWindow::_requestClearQueue( QString user )
{
    Debug::Throw() << "MainWindow::_requestClearQueue - user: " << user << endl;
    if( UserSet::isDefaultUser( user ) ) user = defaultUser_;
    if( user.isEmpty() ) return;

    emit commandRequested( AlienCommand( AlienCommand::ClearQueue ) << _activeFrame().jobManager().user() );
}

//_______________________________________
void MainWindow::_updateConfiguration( void )
{
    Debug::Throw( "MainWindow::_updateConfiguration.\n" );

    // needed to get proper size (from options) at startup
    resize( sizeHint() );

    toggleLogAction_->setChecked( XmlOptions::get().get<bool>( "LOG_ENABLED" ) );
    hideKilledJobsAction_->setChecked( XmlOptions::get().get<bool>( "HIDE_KILLED_JOBS" ) );
    if( XmlOptions::get().get<bool>( "SHOW_COMMAND_LIST" ) ) _showCommandList();
    if( XmlOptions::get().get<bool>( "SHOW_LOG" ) ) _showLog();

    // toggle summary
    toggleSummaryAction_->setChecked( XmlOptions::get().get<bool>( "SHOW_SUMMARY" ) );

    // resubmission
    autoResubmissionEnabled_ = XmlOptions::get().get<bool>( "AUTO_RESUBMISSION_ENABLED" );
    maxResubmissions_ = XmlOptions::get().get<int>( "MAX_RESUBMISSIONS" );

    // refresh rate
    refreshRate_ = 1000 * XmlOptions::get().get<int>( "REFRESH_RATE" );

    // restart timer
    if( interface_->isConnected() && !pauseAction_->isChecked() )
    {
        if( timer_.isActive() ) timer_.stop();
        timer_.start( refreshRate_, this );
    }

}

//________________________________________
void MainWindow::_saveConfiguration( void )
{
    Debug::Throw( "MainWindow::_saveConfiguration.\n" );

    // create job collection
    JobCollection collection;

    // add job sets from existing frames
    for( const auto& frame:Base::KeySet<JobManagerFrame>( this ) )
    {
        // local copies
        JobManager& manager( frame->jobManager() );

        // construct job set
        JobSet jobs( JobSet::fromList( manager.model().children() ) );
        jobs.setOwner( manager.user() );
        jobs.setActive( _isActiveFrame( *frame ) );

        // add to collection
        collection.add( jobs );

    }

    collection.write( File( XmlOptions::get().raw( "RC_FILE" ) ) );

}

//________________________________________
void MainWindow::_installActions( void )
{

    Debug::Throw( "MainWindow::_installActions.\n" );

    // connect
    addAction( connectAction_ = new QAction( IconEngine::get( IconNames::Connect ), tr( "Connect..." ), this ) );
    connect( connectAction_, SIGNAL(triggered()), SLOT(_connect()) );

    // command list
    commandListAction_ = new QAction( IconEngine::get( IconNames::Find ), tr( "Show Command List" ), this );
    connect( commandListAction_, SIGNAL(triggered()), SLOT(_showCommandList()) );

    // log
    logAction_ = new QAction( IconEngine::get( IconNames::Find ), tr( "Show Command Log" ), this );
    logAction_->setEnabled( false );
    connect( logAction_, SIGNAL(triggered()), SLOT(_showLog()) );

    toggleLogAction_ = new QAction( tr( "Enable Command Log" ), this );
    toggleLogAction_->setCheckable( true );
    toggleLogAction_->setChecked( false );
    connect( toggleLogAction_, SIGNAL(toggled(bool)), SLOT(_toggleLog(bool)) );

    // hide killed jobs
    hideKilledJobsAction_ = new QAction( tr( "Hide Killed Jobs" ), this );
    hideKilledJobsAction_->setCheckable( true );
    hideKilledJobsAction_->setChecked( false );
    connect( hideKilledJobsAction_, SIGNAL(toggled(bool)), SLOT(_toggleHideKilledJobs(bool)) );

    // pause
    pauseAction_ = new QAction( IconEngine::get( IconNames::Pause ), tr( "Pause" ), this );
    pauseAction_->setCheckable( true );
    pauseAction_->setChecked( false );
    connect( pauseAction_, SIGNAL(toggled(bool)), this, SLOT(_togglePause(bool)) );

    // default user
    addAction( defaultUserAction_ = new QAction( IconEngine::get( IconNames::Reload ), tr( "Reload Default User" ), this ) );
    connect( defaultUserAction_, SIGNAL(triggered()), SLOT(_requestDefaultUser()) );

    // user list
    addAction( userListAction_ = new QAction( IconEngine::get( IconNames::Reload ), tr( "Reload User List" ), this ) );
    connect( userListAction_, SIGNAL(triggered()), SLOT(_requestUserList()) );

    // update
    addAction( updateAction_ = new QAction( IconEngine::get( IconNames::Reload ), tr( "Reload" ), this ) );
    connect( updateAction_, SIGNAL(triggered()), SLOT(_requestJobList()) );
    updateAction_->setShortcut( QKeySequence::Refresh );
    updateAction_->setShortcutContext( Qt::ApplicationShortcut );

    // kill
    addAction( killAction_ = new QAction( IconEngine::get( IconNames::Kill ), tr( "Kill" ), this ) );
    killAction_->setToolTip( tr( "Kill selected jobs" ) );
    killAction_->setShortcut( Qt::Key_Delete );
    connect( killAction_, SIGNAL(triggered()), SLOT(_kill()) );

    // resubmit
    addAction( resubmitAction_ = new QAction( IconEngine::get( IconNames::Resubmit ), tr( "Resubmit" ), this ) );
    resubmitAction_->setToolTip( tr( "Re-submit failed selected jobs" ) );
    connect( resubmitAction_, SIGNAL(triggered()), SLOT(_resubmit()) );

    // trace
    addAction( traceAction_ = new QAction( IconEngine::get( IconNames::Information ), tr( "Job Properties..." ), this ) );
    traceAction_->setToolTip( tr( "Request information on current job" ) );
    traceAction_->setShortcut( Qt::ALT + Qt::Key_Return );
    connect( traceAction_, SIGNAL(triggered()), SLOT(_requestJobInformation()) );

    // clear
    addAction( clearAction_ = new QAction( IconEngine::get( IconNames::Clear ), tr( "Clear" ), this ) );
    clearAction_->setToolTip( tr( "Remove completed jobs from list" ) );
    clearAction_->setEnabled( false );
    connect( clearAction_, SIGNAL(triggered()), SLOT(_clear()) );

    // abort
    addAction( abortAction_ = new QAction( IconEngine::get( IconNames::Kill ), tr( "Abort" ), this ) );
    abortAction_->setToolTip( tr( "Abort current operations" ) );
    abortAction_->setEnabled( false );

    // summary
    addAction( toggleSummaryAction_ = new QAction( tr( "Show Summary Panel" ), this ) );
    toggleSummaryAction_->setCheckable( true );
    toggleSummaryAction_->setChecked( true );
    connect( toggleSummaryAction_, SIGNAL(toggled(bool)), SLOT(_toggleSummary(bool)) );

    // add frame
    addAction( addFrameAction_ = new QAction( IconEngine::get( IconNames::AddTab ), tr( "Add Tab..." ), this ) );
    addFrameAction_->setShortcut( QKeySequence::AddTab );
    connect( addFrameAction_, SIGNAL(triggered()), SLOT(_addFrame()) );

    // remove frame
    addAction( removeFrameAction_ = new QAction( IconEngine::get( IconNames::RemoveTab ), "Remove Tab", this ) );
    removeFrameAction_->setShortcut( QKeySequence::Close );
    connect( removeFrameAction_, SIGNAL(triggered()), SLOT(_removeFrame()) );
    removeFrameAction_->setEnabled( false );

    // close job dialogs
    addAction( closeJobDialogsAction_ = new QAction( IconEngine::get( IconNames::DialogClose ), "&Close Job Dialogs", this ) );
    connect( closeJobDialogsAction_, SIGNAL(triggered()), SLOT(_closeJobInformationDialogs()) );

}

//________________________________________
void MainWindow::_setActiveFrame( JobManagerFrame& frame )
{
    Debug::Throw( "MainWindow::_setActiveFrame.\n" );
    if( activeFrame_ == &frame ) return;
    activeFrame_ = &frame;

    // update actions
    _updateActions();

}

//________________________________________
JobManagerFrame& MainWindow::_newFrame( void )
{

    Debug::Throw( "MainWindow::_newFrame.\n" );

    JobManagerFrame* frame( new JobManagerFrame() );

    // toggle summary frame
    frame->summaryFrame().setVisible( toggleSummaryAction_->isChecked() );

    // add actions to list
    QMenu* menu( new ContextMenu( &frame->jobManager().list() ) );
    menu->addAction( &frame->jobManager().list().findAction() );
    menu->addAction( &traceAction() );
    menu->addAction( &resubmitAction() );
    menu->addAction( &killAction() );
    menu->addAction( &updateAction() );
    menu->addSeparator();

    // interface connections
    connect( interface_, SIGNAL(jobInformationAvailable(Job,QString)), &frame->jobManager(), SLOT(processJobInformation(Job)) );
    connect( interface_, SIGNAL(childrenInformationAvailable(Job,JobSet)), &frame->jobManager(), SLOT(processChildrenInformation(Job,JobSet)) );


    // list/model connections
    connect( frame->jobManager().list().selectionModel(), SIGNAL(selectionChanged (QItemSelection,QItemSelection)), SLOT(_updateActions()) );
    connect( &frame->jobManager(), SIGNAL(messageAvailable(QString)), SLOT(_updateStatusBar(QString)));
    connect( &frame->jobManager().list(), SIGNAL(expanded(QModelIndex)), SLOT(_updateActions()) );
    connect( &frame->jobManager().list(), SIGNAL(collapsed(QModelIndex)), SLOT(_updateActions()) );
    connect( &frame->jobManager().list(), SIGNAL(activated(QModelIndex)), SLOT(_requestJobInformation()) );

    connect( &frame->jobManager(), SIGNAL(childrenListRequested(Job)), SLOT(_requestChildrenList(Job)) );

    // progressbar connections
    connect( &frame->jobManager(), SIGNAL(busy(int)), statusBar_, SLOT(showProgressBar()) );
    connect( &frame->jobManager(), SIGNAL(busy(int)), &statusBar_->progressBar(), SLOT(setMaximum(int)) );
    connect( &frame->jobManager(), SIGNAL(busy(int)), &statusBar_->progressBar(), SLOT(reset()) );

    connect( &frame->jobManager(), SIGNAL(progressAvailable(int)), &statusBar_->progressBar(), SLOT(setValue(int)) );
    connect( &frame->jobManager(), SIGNAL(idle()), statusBar_, SLOT(showLabel()) );

    // initial focus
    frame->jobManager().list().setFocus();

    // associations
    Base::Key::associate( this, frame );

    // update remove action
    if( Base::KeySet<JobManagerFrame>( this ).size() > 1 )
    { removeFrameAction_->setEnabled( true ); }

    return *frame;
}

//________________________________________
UserSet MainWindow::_users( void ) const
{

    UserSet out;

    // get list of frames
    for( const auto& frame:Base::KeySet<JobManagerFrame>( this ) )
    {
        QString user( frame->jobManager().user() );
        if( UserSet::isDefaultUser( user ) && _hasDefaultUser() ) user = defaultUser_;
        out << user;
    }

    return out;
}

//__________________________________________________________________________
bool MainWindow::_canResubmit( const Job& job ) const
{
    // auto resubmission
    if( !autoResubmissionEnabled_ ) return false;

    // job status
    if( !job.isCompleted() ) return false;

    // user
    if(! (_hasDefaultUser() && job.get( Job::User ) == defaultUser_ ) ) return false;

    // resubmission quota
    if( maxResubmissions_ > 0 && (!job.has( Job::Resubmitted ) || job.get<int>( Job::Resubmitted ) < maxResubmissions_ ) )
    { return false; }

    // all tests passed
    return true;

}

//__________________________________________________________________________
AlienCommand::List MainWindow::_masterJobResubmissionCommands( const Job& job ) const
{
    if( !job.isMasterJob() ) return AlienCommand::List();

    // check failed jobs
    if( job.childrenStatistics().get( JobStatistics::Failed ) <= 0 )
    { return AlienCommand::List(); }

    // job user
    const QString user( job.get( Job::User ) );

    // keep track of failed jobs
    JobModel::List failedJobs;

    // list of states to be checked
    using StringSet = QSet<QString>;
    StringSet statusSet;

    // keep track of whether master job has failed children
    for( const auto& child:_activeFrame().jobManager().model().children( job ) )
    {
        if( child.isFailed() )
        {
            // store status
            statusSet.insert( child.get( Job::Status ) );
            failedJobs.append( child );
        }
    }

    AlienCommand::List commands;
    for( auto status:statusSet )
    {

        // fix status string
        if( status == "Z" ) status = "ZOMBIE";
        else if( !status.isEmpty() && status[0] == 'E' && status.indexOf( "ERROR_" ) < 0 && status != "EXPIRED" )
        { status.replace( QRegExp( "^E" ), "ERROR_" ); }

        // create ResubmitMaster command and append to list
        AlienCommand command( AlienCommand::ResubmitMaster );
        command
            << user
            << QString::number( job.parentId() )
            << QString::number( job.id() )
            << status;
        commands << command;

    }

    // reset timestamps of resubmitted jobs
    for( auto& job:failedJobs )
    {
        job.assign( Job::StartTime, QString() );
        job.assign( Job::EndTime, QString() );
    }

    return commands;

}
